import functools
import inspect
import json
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Text,
    Type,
    TypeVar,
    Union,
)

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.trace import Tracer
from rasa_sdk import Tracker
from rasa_sdk.executor import ActionExecutor, CollectingDispatcher
from rasa_sdk.events import EventType
from rasa_sdk.forms import ValidationAction, FormValidationAction
from rasa_sdk.tracing.instrumentation import attribute_extractors
from rasa_sdk.tracing.tracer_register import ActionExecutorTracerRegister
from rasa_sdk.types import DomainDict


# The `TypeVar` representing the return type for a function to be wrapped.
S = TypeVar("S")
# The `TypeVar` representing the type of the argument passed to the function to be
# wrapped.
T = TypeVar("T")

logger = logging.getLogger(__name__)
INSTRUMENTED_BOOLEAN_ATTRIBUTE_NAME = "class_has_been_instrumented"


def _check_extractor_argument_list(
    fn: Callable[[T, Any, Any], S],
    attr_extractor: Optional[Callable[[T, Any, Any], Dict[str, Any]]],
) -> bool:
    if attr_extractor is None:
        return False

    fn_args = inspect.signature(fn)
    attr_args = inspect.signature(attr_extractor)

    are_arglists_congruent = fn_args.parameters.keys() == attr_args.parameters.keys()

    if not are_arglists_congruent:
        logger.warning(
            f"Argument lists for {fn.__name__} and {attr_extractor.__name__}"
            f" do not match up. {fn.__name__} will be traced without attributes."
        )

    return are_arglists_congruent


def traceable_async(
    fn: Callable[[T, Any, Any], Awaitable[S]],
    tracer: Tracer,
    attr_extractor: Optional[Callable[[T, Any, Any], Dict[str, Any]]],
) -> Callable[[T, Any, Any], Awaitable[S]]:
    """Wrap an `async` function by tracing functionality.

    :param fn: The function to be wrapped.
    :param tracer: The `Tracer` that shall be used for tracing this function.
    :param attr_extractor: A function that is applied to the function's instance and
        the function's arguments.
    :return: The wrapped function.
    """
    should_extract_args = _check_extractor_argument_list(fn, attr_extractor)

    @functools.wraps(fn)
    async def async_wrapper(self: T, *args: Any, **kwargs: Any) -> S:
        attrs = (
            attr_extractor(self, *args, **kwargs)
            if attr_extractor and should_extract_args
            else {}
        )
        if issubclass(self.__class__, FormValidationAction):
            span_name = f"FormValidationAction.{self.__class__.__name__}.{fn.__name__}"
        elif issubclass(self.__class__, ValidationAction):
            span_name = f"ValidationAction.{self.__class__.__name__}.{fn.__name__}"
        else:
            span_name = f"{self.__class__.__name__}.{fn.__name__}"
        with tracer.start_as_current_span(span_name, attributes=attrs):
            return await fn(self, *args, **kwargs)

    return async_wrapper


def traceable(
    fn: Callable[[T, Any, Any], S],
    tracer: Tracer,
    attr_extractor: Optional[Callable[[T, Any, Any], Dict[str, Any]]],
) -> Callable[[T, Any, Any], S]:
    """Wrap a non-`async` function by tracing functionality.

    :param fn: The function to be wrapped.
    :param tracer: The `Tracer` that shall be used for tracing this function.
    :param attr_extractor: A function that is applied to the function's instance and
        the function's arguments.
    :return: The wrapped function.
    """
    should_extract_args = _check_extractor_argument_list(fn, attr_extractor)

    @functools.wraps(fn)
    def wrapper(self: T, *args: Any, **kwargs: Any) -> S:
        # the conditional statement is needed because
        # _create_api_response is a static method
        if isinstance(self, ActionExecutor) and fn.__name__ == "_create_api_response":
            attrs = (
                attr_extractor(*args, **kwargs)
                if attr_extractor and should_extract_args
                else {}
            )
        else:
            attrs = (
                attr_extractor(self, *args, **kwargs)
                if attr_extractor and should_extract_args
                else {}
            )
        with tracer.start_as_current_span(
            f"{self.__class__.__name__}.{fn.__name__}", attributes=attrs
        ):
            if (
                isinstance(self, ActionExecutor)
                and fn.__name__ == "_create_api_response"
            ):
                return fn(*args, **kwargs)
            return fn(self, *args, **kwargs)

    return wrapper


ActionExecutorType = TypeVar("ActionExecutorType", bound=ActionExecutor)
ValidationActionType = TypeVar("ValidationActionType", bound=ValidationAction)
FormValidationActionType = TypeVar(
    "FormValidationActionType", bound=FormValidationAction
)


def instrument(
    tracer_provider: TracerProvider,
    action_executor_class: Optional[Type[ActionExecutorType]] = None,
    validation_action_class: Optional[Type[ValidationActionType]] = None,
    form_validation_action_class: Optional[Type[FormValidationActionType]] = None,
) -> None:
    """Substitute methods to be traced by their traced counterparts.

    :param tracer_provider: The `TracerProvider` to be used for configuring tracing
        on the substituted methods.
    :param action_executor_class: The `ActionExecutor` to be instrumented. If `None`
        is given, no `ActionExecutor` will be instrumented.
    :param validation_action_class: The `ValidationAction` to be instrumented. If `None`
        is given, no `ValidationAction` will be instrumented.
    :param form_validation_action_class: The `FormValidationAction` to be instrumented.
        If `None` is given, no `FormValidationAction` will be instrumented.
    """
    if action_executor_class is not None and not class_is_instrumented(
        action_executor_class
    ):
        tracer = tracer_provider.get_tracer(action_executor_class.__module__)
        _instrument_method(
            tracer,
            action_executor_class,
            "run",
            attribute_extractors.extract_attrs_for_action_executor,
        )
        _instrument_method(
            tracer,
            action_executor_class,
            "_create_api_response",
            attribute_extractors.extract_attrs_for_action_executor_create_api_response,
        )
        mark_class_as_instrumented(action_executor_class)
        ActionExecutorTracerRegister().register_tracer(tracer)

    if validation_action_class is not None and not class_is_instrumented(
        validation_action_class
    ):
        _instrument_method(
            tracer_provider.get_tracer(validation_action_class.__module__),
            validation_action_class,
            "run",
            attribute_extractors.extract_attrs_for_validation_action,
        )
        _instrument_validation_action_extract_validation_events(
            tracer_provider.get_tracer(validation_action_class.__module__),
            validation_action_class,
        )
        mark_class_as_instrumented(validation_action_class)

    if form_validation_action_class is not None and not class_is_instrumented(
        form_validation_action_class
    ):
        _instrument_validation_action_extract_validation_events(
            tracer_provider.get_tracer(form_validation_action_class.__module__),
            form_validation_action_class,
        )
        mark_class_as_instrumented(form_validation_action_class)


def _instrument_method(
    tracer: Tracer,
    instrumented_class: Type,
    method_name: Text,
    attr_extractor: Optional[Callable],
) -> None:
    method_to_trace = getattr(instrumented_class, method_name)
    if inspect.iscoroutinefunction(method_to_trace):
        traced_method = traceable_async(method_to_trace, tracer, attr_extractor)
    else:
        traced_method = traceable(method_to_trace, tracer, attr_extractor)
    setattr(instrumented_class, method_name, traced_method)

    logger.debug(f"Instrumented '{instrumented_class.__name__}.{method_name}'.")


def _mangled_instrumented_boolean_attribute_name(instrumented_class: Type) -> Text:
    # see https://peps.python.org/pep-0008/#method-names-and-instance-variables
    # and https://stackoverflow.com/a/50401073
    return f"_{instrumented_class.__name__}__{INSTRUMENTED_BOOLEAN_ATTRIBUTE_NAME}"


def class_is_instrumented(instrumented_class: Type) -> bool:
    """Check if a class has already been instrumented."""
    return getattr(
        instrumented_class,
        _mangled_instrumented_boolean_attribute_name(instrumented_class),
        False,
    )


def mark_class_as_instrumented(instrumented_class: Type) -> None:
    """Mark a class as instrumented if it isn't already marked."""
    if not class_is_instrumented(instrumented_class):
        setattr(
            instrumented_class,
            _mangled_instrumented_boolean_attribute_name(instrumented_class),
            True,
        )


def _instrument_validation_action_extract_validation_events(
    tracer: Tracer,
    validation_action_class: Union[Type[ValidationAction], Type[FormValidationAction]],
) -> None:
    def tracing_validation_action_extract_validation_events_wrapper(
        fn: Callable,
    ) -> Callable:
        @functools.wraps(fn)
        async def wrapper(
            self,
            dispatcher: CollectingDispatcher,
            tracker: Tracker,
            domain: DomainDict,
        ) -> List[EventType]:
            if issubclass(self.__class__, FormValidationAction):
                span_name = (
                    f"FormValidationAction.{self.__class__.__name__}.{fn.__name__}"
                )
            else:
                span_name = f"ValidationAction.{self.__class__.__name__}.{fn.__name__}"

            with tracer.start_as_current_span(span_name) as span:
                validation_events = await fn(self, dispatcher, tracker, domain)
                event_names = []
                slot_names = []

                if validation_events:
                    for event in validation_events:
                        event_names.append(event.get("event"))
                        if event.get("event") == "slot":
                            slot_names.append(event.get("name"))

                span.set_attributes(
                    {
                        "validation_events": json.dumps(
                            list(dict.fromkeys(event_names))
                        ),
                        "slots": json.dumps(list(dict.fromkeys(slot_names))),
                    }
                )
                return validation_events

        return wrapper

    validation_action_class._extract_validation_events = (  # type: ignore
        tracing_validation_action_extract_validation_events_wrapper(
            validation_action_class._extract_validation_events
        )
    )

    logger.debug(
        f"Instrumented '{validation_action_class.__name__}._extract_validation_events'."
    )



import logging
import typing
import warnings
from typing import Dict, Text, Any, List, Optional

from abc import ABC
from rasa_sdk import utils
from rasa_sdk.events import SlotSet, EventType
from rasa_sdk.interfaces import Action

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:  # pragma: no cover
    from rasa_sdk import Tracker
    from rasa_sdk.executor import CollectingDispatcher
    from rasa_sdk.types import DomainDict

# this slot is used to store information needed
# to do the form handling
REQUESTED_SLOT = "requested_slot"

LOOP_INTERRUPTED_KEY = "is_interrupted"

ACTION_VALIDATE_SLOT_MAPPINGS = "action_validate_slot_mappings"


class ValidationAction(Action, ABC):
    """A helper class for slot validations and extractions of custom slots."""

    def name(self) -> Text:
        """Unique identifier of this simple action."""
        return ACTION_VALIDATE_SLOT_MAPPINGS

    async def run(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[EventType]:
        """Runs the custom actions. Please the docstring of the parent class."""
        extraction_events = await self.get_extraction_events(
            dispatcher, tracker, domain
        )
        tracker.add_slots(extraction_events)

        validation_events = await self._extract_validation_events(
            dispatcher, tracker, domain
        )

        # Validation events include events for extracted slots
        return validation_events

    async def _extract_validation_events(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[EventType]:
        validation_events = await self.get_validation_events(
            dispatcher, tracker, domain
        )
        tracker.add_slots(validation_events)

        return validation_events

    async def required_slots(
        self,
        domain_slots: List[Text],
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[Text]:
        """Returns slots which the validation action should fill.

        Args:
            domain_slots: Names of slots of this form which were mapped in
                the domain.
            dispatcher: the dispatcher which is used to
                send messages back to the user.
            tracker: the conversation tracker for the current user.
            domain: the bot's domain.

        Returns:
            A list of slot names.
        """
        return domain_slots

    async def get_extraction_events(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[EventType]:
        """Extracts custom slots using available `extract_<slot name>` methods.

        Uses the information from `self.required_slots` to gather which slots should
        be extracted.

        Args:
            dispatcher: the dispatcher which is used to
                send messages back to the user. Use
                `dispatcher.utter_message()` for sending messages.
            tracker: the state tracker for the current
                user. You can access slot values using
                `tracker.get_slot(slot_name)`, the most recent user message
                is `tracker.latest_message.text` and any other
                `rasa_sdk.Tracker` property.
            domain: the bot's domain.

        Returns:
            `SlotSet` for any extracted slots.
        """
        custom_slots = {}
        slots_to_extract = await self.required_slots(
            self.domain_slots(domain), dispatcher, tracker, domain
        )

        for slot in slots_to_extract:
            extraction_output = await self._extract_slot(
                slot, dispatcher, tracker, domain
            )
            custom_slots.update(extraction_output)
            # for sequential consistency, also update tracker
            # to make changes visible to subsequent extract_{slot_name}
            tracker.slots.update(extraction_output)

        return [SlotSet(slot, value) for slot, value in custom_slots.items()]

    async def get_validation_events(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[EventType]:
        """Validate slots by calling a validation function for each slot.

        Args:
            dispatcher: the dispatcher which is used to
                send messages back to the user.
            tracker: the conversation tracker for the current user.
            domain: the bot's domain.

        Returns:
            `SlotSet` events for every validated slot.
        """
        slots_to_validate = await self.required_slots(
            self.domain_slots(domain), dispatcher, tracker, domain
        )
        slots: Dict[Text, Any] = tracker.slots_to_validate()

        for slot_name, slot_value in list(slots.items()):
            if slot_name not in slots_to_validate:
                slots.pop(slot_name)
                continue

            method_name = f"validate_{slot_name.replace('-','_')}"
            validate_method = getattr(self, method_name, None)

            if not validate_method:
                logger.warning(
                    f"Skipping validation for `{slot_name}`: there is no validation "
                    f"method specified."
                )
                continue

            validation_output = await utils.call_potential_coroutine(
                validate_method(slot_value, dispatcher, tracker, domain)
            )

            if isinstance(validation_output, dict):
                slots.update(validation_output)
                # for sequential consistency, also update tracker
                # to make changes visible to subsequent validate_{slot_name}
                tracker.slots.update(validation_output)
            else:
                warnings.warn(
                    f"Cannot validate `{slot_name}`: make sure the validation method "
                    f"returns the correct output."
                )

        return [SlotSet(slot, value) for slot, value in slots.items()]

    @staticmethod
    def _is_mapped_to_form(slot_value: Dict[Text, Any]) -> bool:
        mappings = slot_value.get("mappings")
        if not mappings:
            return False

        for mapping in mappings:
            mapping_conditions = mapping.get("conditions", [])
            for condition in mapping_conditions:
                if condition.get("active_loop"):
                    return True

        return False

    def global_slots(self, domain: "DomainDict") -> List[Text]:
        """Returns all slots that contain no form condition."""
        all_slots = domain.get("slots", {})
        return [k for k, v in all_slots.items() if not self._is_mapped_to_form(v)]

    def domain_slots(self, domain: "DomainDict") -> List[Text]:
        """Returns slots which were mapped in the domain.

        Args:
            domain: The current domain.

        Returns:
            Slot names mapped in the domain which do not include
            a mapping with an active loop condition.
        """
        return self.global_slots(domain)

    async def _extract_slot(
        self,
        slot_name: Text,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> Dict[Text, Any]:
        method_name = f"extract_{slot_name.replace('-', '_')}"

        slot_in_domain = slot_name in self.domain_slots(domain)
        extract_method = getattr(self, method_name, None)

        if not extract_method:
            if not slot_in_domain:
                warnings.warn(
                    f"No method '{method_name}' found for slot "
                    f"'{slot_name}'. Skipping extraction for this slot."
                )
            return {}

        extracted = await utils.call_potential_coroutine(
            extract_method(dispatcher, tracker, domain)
        )

        if isinstance(extracted, dict):
            return extracted

        warnings.warn(
            f"Cannot extract `{slot_name}`: make sure the extract method "
            f"returns the correct output."
        )
        return {}


class FormValidationAction(ValidationAction, ABC):
    """A helper class for slot validations and extractions of custom slots in forms."""

    def name(self) -> Text:
        """Unique identifier of this simple action."""
        raise NotImplementedError("An action must implement a name")

    def form_name(self) -> Text:
        """Returns the form's name."""
        return self.name().replace("validate_", "", 1)

    async def _extract_validation_events(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> List[EventType]:
        validation_events = await self.get_validation_events(
            dispatcher, tracker, domain
        )
        tracker.add_slots(validation_events)

        next_slot = await self.next_requested_slot(dispatcher, tracker, domain)
        if next_slot:
            validation_events.append(next_slot)

        return validation_events

    def domain_slots(self, domain: "DomainDict") -> List[Text]:
        """Returns slots which were mapped in the domain.

        Args:
            domain: The current domain.

        Returns:
            Slot names which should be filled by the form. By default it
            returns the slot names which are listed for this form in the domain
            and use predefined mappings.
        """
        form = domain.get("forms", {}).get(self.form_name(), {})
        if "required_slots" in form:
            return form.get("required_slots", [])
        return []

    async def next_requested_slot(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: "Tracker",
        domain: "DomainDict",
    ) -> Optional[EventType]:
        """Sets the next slot which should be requested.

        Skips setting the next requested slot in case `missing_slots` was not
        overridden.

        Args:
            dispatcher: the dispatcher which is used to
                send messages back to the user.
            tracker: the conversation tracker for the current user.
            domain: the bot's domain.

        Returns:
            `None` in case `missing_slots` was not overridden and returns `None`.
            Otherwise returns a `SlotSet` event for the next slot to be requested.
            If the `SlotSet` event sets `requested_slot` to `None`, the form will be
            deactivated.
        """
        required_slots = await self.required_slots(
            self.domain_slots(domain), dispatcher, tracker, domain
        )
        if required_slots == self.domain_slots(domain):
            # If users didn't override `required_slots` then we'll let the `FormAction`
            # within Rasa Open Source request the next slot.
            return None

        missing_slots = (
            slot_name
            for slot_name in required_slots
            if tracker.slots.get(slot_name) is None
        )

        return SlotSet(REQUESTED_SLOT, next(missing_slots, None))



import logging

import os
from typing import Any, Dict, Optional, Text
import rasa_sdk.utils


logger = logging.getLogger(__name__)
DEFAULT_ENCODING = "utf-8"


def read_endpoint_config(
    filename: Text, endpoint_type: Text
) -> Optional["EndpointConfig"]:
    """Read an endpoint configuration file from disk and extract one config.

    Args:
        filename: the endpoint config file to read
        endpoint_type: the type of the endpoint

    Returns:
        The endpoint configuration of the passed type if it exists, `None` otherwise.
    """
    if not filename:
        return None

    try:
        content = rasa_sdk.utils.read_file(filename)
        content = rasa_sdk.utils.read_yaml(content)

        if content.get(endpoint_type) is None:
            return None

        return EndpointConfig.from_dict(content[endpoint_type])
    except FileNotFoundError:
        logger.error(
            "Failed to read endpoint configuration " "from {}. No such file.".format(
                os.path.abspath(filename)
            )
        )
        return None


class EndpointConfig:
    """Configuration for an external HTTP endpoint."""

    def __init__(
        self,
        url: Optional[Text] = None,
        params: Optional[Dict[Text, Any]] = None,
        headers: Optional[Dict[Text, Any]] = None,
        basic_auth: Optional[Dict[Text, Text]] = None,
        token: Optional[Text] = None,
        token_name: Text = "token",
        cafile: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        """Creates an `EndpointConfig` instance."""
        self.url = url
        self.params = params or {}
        self.headers = headers or {}
        self.basic_auth = basic_auth or {}
        self.token = token
        self.token_name = token_name
        self.type = kwargs.pop("store_type", kwargs.pop("type", None))
        self.cafile = cafile
        self.kwargs = kwargs

    @classmethod
    def from_dict(cls, data: Dict[Text, Any]) -> "EndpointConfig":
        return EndpointConfig(**data)



import asyncio
import inspect
import logging
import logging.config
import warnings
import os
from pathlib import Path
from ruamel import yaml as yaml
from ruamel.yaml import YAMLError
from ruamel.yaml.constructor import DuplicateKeyError

from typing import (
    AbstractSet,
    Any,
    ClassVar,
    Dict,
    List,
    Text,
    Optional,
    Coroutine,
    Union,
)

import rasa_sdk

from rasa_sdk.constants import (
    DEFAULT_ENCODING,
    DEFAULT_SANIC_WORKERS,
    ENV_SANIC_WORKERS,
    DEFAULT_LOG_LEVEL_LIBRARIES,
    ENV_LOG_LEVEL_LIBRARIES,
    PYTHON_LOGGING_SCHEMA_DOCS,
    YAML_VERSION,
)
from rasa_sdk.exceptions import (
    FileIOException,
    FileNotFoundException,
    YamlSyntaxException,
)

logger = logging.getLogger(__name__)


class Element(dict):
    """Represents an element in a list of elements in a rich message."""

    __acceptable_keys: ClassVar[List[Text]] = [
        "title",
        "item_url",
        "image_url",
        "subtitle",
        "buttons",
    ]

    def __init__(self, *args, **kwargs):
        """Initializes an element in a list of elements in a rich message."""
        kwargs = {
            key: value for key, value in kwargs.items() if key in self.__acceptable_keys
        }

        super().__init__(*args, **kwargs)


class Button(dict):
    """Represents a button in a rich message."""

    pass


class Singleton(type):
    """Singleton metaclass."""

    _instances: ClassVar[Dict[Any, Any]] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        """Call the class.

        Args:
            *args: Arguments.
            **kwargs: Keyword arguments.
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]

    @classmethod
    def clear(cls) -> None:
        """Clear the class."""
        cls._instances = {}


def all_subclasses(cls: Any) -> List[Any]:
    """Returns all known (imported) subclasses of a class."""
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in all_subclasses(s)
    ]


def add_logging_level_option_arguments(parser):
    """Add options to an argument parser to configure logging levels."""
    # arguments for logging configuration
    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose. Sets logging level to INFO",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
        default=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--debug",
        help="Print lots of debugging statements. Sets logging level to DEBUG",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
    )
    parser.add_argument(
        "--quiet",
        help="Be quiet! Sets logging level to WARNING",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
    )


def add_logging_file_arguments(parser):
    """Add options to an argument parser to configure logging to a file."""
    parser.add_argument(
        "--log-file",
        type=str,
        default=None,
        help="Store logs in specified file.",
    )
    parser.add_argument(
        "--logging-config_file",
        type=str,
        default=None,
        help="If set, the name of the logging configuration file will be set "
        "to the given name.",
    )


def configure_colored_logging(loglevel):
    """Configure logging with colors."""
    import coloredlogs

    field_styles = coloredlogs.DEFAULT_FIELD_STYLES.copy()
    field_styles["asctime"] = {}
    level_styles = coloredlogs.DEFAULT_LEVEL_STYLES.copy()
    level_styles["debug"] = {}
    coloredlogs.install(
        level=loglevel,
        use_chroot=False,
        fmt="%(asctime)s %(levelname)-8s %(name)s  - %(message)s",
        level_styles=level_styles,
        field_styles=field_styles,
    )


def configure_logging_from_input_file(logging_config_file: Union[Path, Text]) -> None:
    """Parses YAML file content to configure logging.

    Args:
        logging_config_file: YAML file containing logging configuration to handle
            custom formatting
    """
    logging_config_dict = read_yaml_file(logging_config_file)

    try:
        logging.config.dictConfig(logging_config_dict)
    except (ValueError, TypeError, AttributeError, ImportError) as e:
        logging.debug(
            f"The logging config file {logging_config_file} could not "
            f"be applied because it failed validation against "
            f"the built-in Python logging schema. "
            f"More info at {PYTHON_LOGGING_SCHEMA_DOCS}.",
            exc_info=e,
        )


def set_default_logging(
    logger_obj: logging.Logger, output_log_file: Optional[Text], loglevel: int
) -> None:
    """Configure default logging to a file.

    :param logger_obj: Logger object to configure.
    :param output_log_file: Path of log file to write to.
    :param loglevel: Log Level.
    :return: None.
    """
    if not output_log_file:
        return

    if not loglevel:
        loglevel = logging.INFO

    logger_obj.setLevel(loglevel)

    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)-5.5s]  %(name)s  -  %(message)s"
    )
    file_handler = logging.FileHandler(output_log_file, encoding=DEFAULT_ENCODING)
    file_handler.setLevel(loglevel)
    file_handler.setFormatter(formatter)
    logger_obj.addHandler(file_handler)


def configure_file_logging(
    logger_obj: logging.Logger,
    output_log_file: Optional[Text],
    loglevel: int,
    logging_config_file: Optional[Text],
) -> None:
    """Configure logging configuration.

    :param logger_obj: Logger object to configure.
    :param output_log_file: Path of log file to write to.
    :param loglevel: Log Level.
    :param logging_config_file: YAML file containing logging configuration to handle
            custom formatting
    :return: None.
    """
    if logging_config_file is not None:
        configure_logging_from_input_file(logging_config_file)
        return

    set_default_logging(logger_obj, output_log_file, loglevel)


def arguments_of(func) -> AbstractSet[Text]:
    """Return the parameters of the function `func` as a list of their names."""
    return inspect.signature(func).parameters.keys()


def number_of_sanic_workers() -> int:
    """Get the number of Sanic workers to use in `app.run()`.

    If the environment variable `constants.ENV_SANIC_WORKERS`
    is set and is not equal to 1.
    """

    def _log_and_get_default_number_of_workers():
        logger.debug(
            f"Using the default number of Sanic workers ({DEFAULT_SANIC_WORKERS})."
        )
        return DEFAULT_SANIC_WORKERS

    try:
        env_value = int(os.environ.get(ENV_SANIC_WORKERS, DEFAULT_SANIC_WORKERS))
    except ValueError:
        logger.error(
            f"Cannot convert environment variable `{ENV_SANIC_WORKERS}` "
            f"to int ('{os.environ[ENV_SANIC_WORKERS]}')."
        )
        return _log_and_get_default_number_of_workers()

    if env_value == DEFAULT_SANIC_WORKERS:
        return _log_and_get_default_number_of_workers()

    if env_value < 1:
        warnings.warn(
            f"Cannot set number of Sanic workers to the desired value "
            f"({env_value}). The number of workers must be at least 1."
        )
        return _log_and_get_default_number_of_workers()

    logger.debug(f"Using {env_value} Sanic workers.")
    return env_value


def check_version_compatibility(rasa_version: Optional[Text]) -> None:
    """Check if the version of rasa and rasa_sdk are compatible.

    The version check relies on the version string being formatted as
    'x.y.z' and compares whether the numbers x and y are the same for both
    rasa and rasa_sdk.

    Args:
        rasa_version: A string containing the version of rasa that
                        is making the call to the action server.

    Raises:
        Warning: The version of rasa version unknown or not compatible with
                    this version of rasa_sdk.
    """
    # Check for versions of Rasa that are too old to report their version number
    if rasa_version is None:
        warnings.warn(
            f"You are using an old version of rasa which might "
            f"not be compatible with this version of rasa_sdk "
            f"({rasa_sdk.__version__}).\n"
            f"To ensure compatibility use the same version "
            f"for both, modulo the last number, i.e. using version "
            f"A.B.x the numbers A and B should be identical for "
            f"both rasa and rasa_sdk."
        )
        return

    rasa = rasa_version.split(".")[:-1]
    sdk = rasa_sdk.__version__.split(".")[:-1]

    if rasa != sdk:
        warnings.warn(
            f"Your versions of rasa and "
            f"rasa_sdk might not be compatible. You "
            f"are currently running rasa version {rasa_version} "
            f"and rasa_sdk version {rasa_sdk.__version__}.\n"
            f"To ensure compatibility use the same "
            f"version for both, modulo the last number, "
            f"i.e. using version A.B.x the numbers A and "
            f"B should be identical for "
            f"both rasa and rasa_sdk."
        )


def update_sanic_log_level() -> None:
    """Set the log level of sanic loggers.

    Use the environment variable 'LOG_LEVEL_LIBRARIES', or default to
    `DEFAULT_LOG_LEVEL_LIBRARIES` if undefined.
    """
    from sanic.log import logger, error_logger, access_logger

    log_level = os.environ.get(ENV_LOG_LEVEL_LIBRARIES, DEFAULT_LOG_LEVEL_LIBRARIES)

    logger.setLevel(log_level)
    error_logger.setLevel(log_level)
    access_logger.setLevel(log_level)

    logger.propagate = False
    error_logger.propagate = False
    access_logger.propagate = False


async def call_potential_coroutine(
    coroutine_or_return_value: Union[Any, Coroutine],
) -> Any:
    """Await if it's a coroutine."""
    if asyncio.iscoroutine(coroutine_or_return_value):
        return await coroutine_or_return_value

    return coroutine_or_return_value


def read_file(filename: Union[Text, Path], encoding: Text = DEFAULT_ENCODING) -> Any:
    """Read text from a file."""
    try:
        with open(filename, encoding=encoding) as f:
            return f.read()
    except FileNotFoundError:
        raise FileNotFoundException(
            f"Failed to read file, " f"'{os.path.abspath(filename)}' does not exist."
        )
    except UnicodeDecodeError:
        raise FileIOException(
            f"Failed to read file '{os.path.abspath(filename)}', "
            f"could not read the file using {encoding} to decode "
            f"it. Please make sure the file is stored with this "
            f"encoding."
        )


def read_yaml(content: Text, reader_type: Text = "safe") -> Any:
    """Parses yaml from a text.

    Args:
        content: A text containing yaml content.
        reader_type: Reader type to use. By default, "safe" will be used.

    Raises:
        ruamel.yaml.parser.ParserError: If there was an error when parsing the YAML.
    """
    if _is_ascii(content):
        # Required to make sure emojis are correctly parsed
        content = (
            content.encode("utf-8")
            .decode("raw_unicode_escape")
            .encode("utf-16", "surrogatepass")
            .decode("utf-16")
        )

    yaml_parser = yaml.YAML(typ=reader_type)
    yaml_parser.version = YAML_VERSION
    yaml_parser.preserve_quotes = True

    return yaml_parser.load(content) or {}


def _is_ascii(text: Text) -> bool:
    return all(ord(character) < 128 for character in text)


def read_yaml_file(filename: Union[Text, Path]) -> Dict[Text, Any]:
    """Parses a yaml file.

    Raises an exception if the content of the file can not be parsed as YAML.

    Args:
        filename: The path to the file which should be read.

    Returns:
        Parsed content of the file.
    """
    try:
        return read_yaml(read_file(filename, DEFAULT_ENCODING))
    except (YAMLError, DuplicateKeyError) as e:
        raise YamlSyntaxException(filename, e)


def file_as_bytes(file_path: Text) -> bytes:
    """Read in a file as a byte array.

    Args:
        file_path: Path to the file to read.

    Returns:
        The file content as a byte array.

    Raises:
        FileNotFoundException: If the file does not exist.
    """
    try:
        with open(file_path, "rb") as f:
            return f.read()
    except FileNotFoundError:
        raise FileNotFoundException(
            f"Failed to read file, " f"'{os.path.abspath(file_path)}' does not exist."
        )



import functools
import sys
import logging
import pluggy

from sanic import Sanic

hookspec = pluggy.HookspecMarker("rasa_sdk")
logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=2)
def plugin_manager() -> pluggy.PluginManager:
    """Initialises a plugin manager which registers hook implementations."""
    _plugin_manager = pluggy.PluginManager("rasa_sdk")
    _plugin_manager.add_hookspecs(sys.modules["rasa_sdk.plugin"])
    _discover_plugins(_plugin_manager)

    return _plugin_manager


def _discover_plugins(manager: pluggy.PluginManager) -> None:
    try:
        # rasa_sdk_plugin is a custom package
        # which extends existing functionality on rasa action server via plugins
        import rasa_sdk_plugins

        rasa_sdk_plugins.init_hooks(manager)
    except ModuleNotFoundError as e:
        logger.debug("No plugins found: %s", e)
        pass


@hookspec
def attach_sanic_app_extensions(app: Sanic) -> None:
    """Hook specification for attaching sanic listeners, routes and middlewares."""



from multidict import MultiDict

from rasa_sdk.tracing import config
from opentelemetry import trace
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from opentelemetry.sdk.trace import TracerProvider

from typing import Optional, Tuple, Any


def get_tracer_provider(endpoints_file: str) -> Optional[TracerProvider]:
    """Gets the tracer provider from the command line arguments."""
    tracer_provider = config.get_tracer_provider(endpoints_file)
    config.configure_tracing(tracer_provider)

    return tracer_provider


def get_tracer_and_context(
    span_name: str,
    tracer_provider: Optional[TracerProvider],
    tracing_carrier: Optional[MultiDict],
) -> Tuple[Any, Any]:
    """Gets tracer and context."""
    if tracer_provider is None:
        tracer = trace.get_tracer(span_name)
        context = None
    else:
        tracer = tracer_provider.get_tracer(span_name)
        context = (
            TraceContextTextMapPropagator().extract(tracing_carrier)
            if tracing_carrier
            else None
        )
    return tracer, context


def set_span_attributes(span: Any, action_call: dict) -> None:
    """Sets span attributes."""
    tracker = action_call.get("tracker", {})
    span_attributes = {
        "next_action": action_call.get("next_action"),
        "version": action_call.get("version"),
        "sender_id": tracker.get("sender_id"),
        "message_id": tracker.get("latest_message", {}).get("message_id", "None"),
    }

    if span.is_recording():
        for key, value in span_attributes.items():
            span.set_attribute(key, value)

    return None



import argparse

from rasa_sdk.constants import DEFAULT_SERVER_PORT, DEFAULT_ENDPOINTS_PATH


def action_arg(actions_module_path: str) -> str:
    """Validate the action module path.

    Valid action module path is python module, so it should not contain a slash.

    Args:
        actions_module_path: Path to the actions python module.

    Returns:
        actions_module_path: If provided module path is valid.

    Raises:
        argparse.ArgumentTypeError: If the module path is invalid.
    """
    if "/" in actions_module_path:
        raise argparse.ArgumentTypeError(
            "Invalid actions format. Actions file should be a python module "
            "and passed with module notation (e.g. directory.actions)."
        )
    else:
        return actions_module_path


def add_endpoint_arguments(parser: argparse.ArgumentParser) -> None:
    """Add all the arguments to the argument parser."""
    parser.add_argument(
        "-p",
        "--port",
        default=DEFAULT_SERVER_PORT,
        type=int,
        help="port to run the server at",
    )
    parser.add_argument(
        "--cors",
        nargs="*",
        type=str,
        help="enable CORS for the passed origin. Use * to whitelist all origins",
    )
    parser.add_argument(
        "--actions",
        type=action_arg,
        default=None,
        help="name of action package to be loaded",
    )
    parser.add_argument(
        "--actions-module",
        type=action_arg,
        default=None,
        help="name of action package to be loaded",
    )
    parser.add_argument(
        "--ssl-keyfile",
        default=None,
        help="Set the SSL certificate to create a TLS secured server.",
    )
    parser.add_argument(
        "--ssl-certificate",
        default=None,
        help="Set the SSL certificate to create a TLS secured server.",
    )
    parser.add_argument(
        "--ssl-password",
        default=None,
        help="If your ssl-keyfile is protected by a password, you can specify it "
        "using this parameter. "
        "Not supported in grpc mode.",
    )
    parser.add_argument(
        "--ssl-ca-file",
        default=None,
        help="If you want to authenticate the client using a certificate, you can "
        "specify the CA certificate of the client using this parameter. "
        "Supported only in grpc mode.",
    )
    parser.add_argument(
        "--auto-reload",
        help="Enable auto-reloading of modules containing Action subclasses.",
        action="store_true",
    )
    parser.add_argument(
        "--endpoints",
        default=DEFAULT_ENDPOINTS_PATH,
        help="Configuration file for the assistant as a yml file.",
    )
    parser.add_argument(
        "--grpc", help="Starts grpc server instead of http", action="store_true"
    )



from __future__ import annotations

import signal

import asyncio

import grpc
import logging
from typing import Optional, Any, Dict
from concurrent import futures
from grpc import aio
from grpc_health.v1 import health
from grpc_health.v1 import health_pb2
from grpc_health.v1 import health_pb2_grpc
from google.protobuf.json_format import MessageToDict, ParseDict
from grpc.aio import Metadata
from multidict import MultiDict

from rasa_sdk.constants import (
    DEFAULT_SERVER_PORT,
    DEFAULT_ENDPOINTS_PATH,
    NO_GRACE_PERIOD,
)
from rasa_sdk.executor import ActionExecutor
from rasa_sdk.grpc_errors import (
    ResourceNotFound,
    ResourceNotFoundType,
    ActionExecutionFailed,
)
from rasa_sdk.grpc_py import (
    action_webhook_pb2,
    action_webhook_pb2_grpc,
)
from rasa_sdk.grpc_py.action_webhook_pb2 import (
    ActionsResponse,
    ActionsRequest,
    WebhookRequest,
)
from rasa_sdk.interfaces import (
    ActionExecutionRejection,
    ActionNotFoundException,
    ActionMissingDomainException,
)
from rasa_sdk.tracing.utils import (
    get_tracer_provider,
    TracerProvider,
    get_tracer_and_context,
    set_span_attributes,
)
from rasa_sdk.utils import (
    check_version_compatibility,
    number_of_sanic_workers,
    file_as_bytes,
)

logger = logging.getLogger(__name__)


GRPC_ACTION_SERVER_NAME = "ActionServer"


class GRPCActionServerWebhook(action_webhook_pb2_grpc.ActionServiceServicer):
    """Runs webhook RPC which is served through gRPC server."""

    def __init__(
        self,
        executor: ActionExecutor,
        auto_reload: bool = False,
        tracer_provider: Optional[TracerProvider] = None,
    ) -> None:
        """Initializes the ActionServerWebhook.

        Args:
            tracer_provider: The tracer provider.
            auto_reload: Enable auto-reloading of modules containing Action subclasses.
            executor: The action executor.
        """
        self.tracer_provider = tracer_provider
        self.auto_reload = auto_reload
        self.executor = executor

    async def Actions(
        self,
        request: ActionsRequest,
        context: grpc.aio.ServicerContext,
    ) -> ActionsResponse:
        """Handle RPC request for the actions.

        Args:
            request: The actions request.
            context: The context of the request.

        Returns:
            gRPC response.
        """
        if self.auto_reload:
            self.executor.reload()

        actions = [action.model_dump() for action in self.executor.list_actions()]
        response = ActionsResponse()
        return ParseDict(
            {
                "actions": actions,
            },
            response,
        )

    async def Webhook(
        self,
        request: WebhookRequest,
        context: grpc.aio.ServicerContext,
    ) -> action_webhook_pb2.WebhookResponse:
        """Handle RPC request for the webhook.

        Args:
            request: The webhook request.
            context: The context of the request.

        Returns:
            gRPC response.
        """
        span_name = "GRPCActionServerWebhook.Webhook"
        invocation_metadata = context.invocation_metadata()

        def convert_metadata_to_multidict(
            metadata: Optional[Metadata],
        ) -> Optional[MultiDict]:
            """Convert list of tuples to multidict."""
            if not metadata:
                return None
            return MultiDict(metadata)

        tracer, tracing_context = get_tracer_and_context(
            span_name=span_name,
            tracer_provider=self.tracer_provider,
            tracing_carrier=convert_metadata_to_multidict(invocation_metadata),
        )

        with tracer.start_as_current_span(span_name, context=tracing_context) as span:
            check_version_compatibility(request.version)
            if self.auto_reload:
                self.executor.reload()
            try:
                action_call = MessageToDict(request, preserving_proto_field_name=True)
                result = await self.executor.run(action_call)
            except ActionExecutionRejection as e:
                logger.debug(e)

                body = ActionExecutionFailed(
                    action_name=e.action_name, message=e.message
                ).model_dump_json()
                context.set_code(grpc.StatusCode.INTERNAL)
                context.set_details(body)
                return action_webhook_pb2.WebhookResponse()
            except ActionNotFoundException as e:
                logger.error(e)
                body = ResourceNotFound(
                    action_name=e.action_name,
                    message=e.message,
                    resource_type=ResourceNotFoundType.ACTION,
                ).model_dump_json()
                context.set_code(grpc.StatusCode.NOT_FOUND)
                context.set_details(body)
                return action_webhook_pb2.WebhookResponse()
            except ActionMissingDomainException as e:
                logger.error(e)
                body = ResourceNotFound(
                    action_name=e.action_name,
                    message=e.message,
                    resource_type=ResourceNotFoundType.DOMAIN,
                ).model_dump_json()
                context.set_code(grpc.StatusCode.NOT_FOUND)
                context.set_details(body)
                return action_webhook_pb2.WebhookResponse()
            if not result:
                return action_webhook_pb2.WebhookResponse()

            _set_grpc_span_attributes(span, action_call, method_name="Webhook")
            response = action_webhook_pb2.WebhookResponse()
            return ParseDict(result.model_dump(), response)


def _set_grpc_span_attributes(
    span: Any, action_call: Dict[str, Any], method_name: str
) -> None:
    """Sets grpc span attributes."""
    set_span_attributes(span, action_call)
    if span.is_recording():
        span.set_attribute("grpc.method", method_name)


def _get_signal_name(signal_number: int) -> str:
    """Return the signal name for the given signal number."""
    return signal.Signals(signal_number).name


def _initialise_interrupts(server: grpc.Server) -> None:
    """Initialise handlers for kernel signal interrupts."""

    async def handle_sigint(signal_received: int):
        """Handle the received signal."""
        logger.info(
            f"Received {_get_signal_name(signal_received)} signal."
            "Stopping gRPC server..."
        )
        await server.stop(NO_GRACE_PERIOD)
        logger.info("gRPC server stopped.")

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(
        signal.SIGINT, lambda: asyncio.create_task(handle_sigint(signal.SIGINT))
    )
    loop.add_signal_handler(
        signal.SIGTERM, lambda: asyncio.create_task(handle_sigint(signal.SIGTERM))
    )


def _initialise_health_service(server: grpc.Server):
    """Initialise the health service.

    Args:
        server: The gRPC server.
    """
    health_servicer = health.HealthServicer(
        experimental_non_blocking=True,
        experimental_thread_pool=futures.ThreadPoolExecutor(max_workers=10),
    )
    health_servicer.set(GRPC_ACTION_SERVER_NAME, health_pb2.HealthCheckResponse.SERVING)
    health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)


def _initialise_action_service(
    server: grpc.Server,
    action_executor: ActionExecutor,
    auto_reload: bool,
    endpoints: str,
):
    """Initialise the action service.

    Args:
        server: The gRPC server.
        action_executor: The action executor.
        auto_reload: Enable auto-reloading of modules containing Action subclasses.
        endpoints: Path to the endpoints file.
    """
    tracer_provider = get_tracer_provider(endpoints)
    action_webhook_pb2_grpc.add_ActionServiceServicer_to_server(
        GRPCActionServerWebhook(action_executor, auto_reload, tracer_provider), server
    )


def _initialise_port(
    server: grpc.Server,
    port: int = DEFAULT_SERVER_PORT,
    ssl_server_cert: Optional[bytes] = None,
    ssl_server_cert_key: Optional[bytes] = None,
    ssl_ca_cert: Optional[bytes] = None,
) -> None:
    if ssl_server_cert and ssl_server_cert_key:
        # Use SSL/TLS if certificate and key are provided
        grpc.ssl_channel_credentials()
        logger.info(f"Starting gRPC server with SSL support on port {port}")
        server.add_secure_port(
            f"[::]:{port}",
            server_credentials=grpc.ssl_server_credentials(
                private_key_certificate_chain_pairs=[
                    (ssl_server_cert_key, ssl_server_cert)
                ],
                root_certificates=ssl_ca_cert,
                require_client_auth=True if ssl_ca_cert else False,
            ),
        )
    else:
        logger.info(f"Starting gRPC server without SSL on port {port}")
        # Use insecure connection if no SSL/TLS information is provided
        server.add_insecure_port(f"[::]:{port}")


def _initialise_grpc_server(
    action_executor: ActionExecutor,
    port: int = DEFAULT_SERVER_PORT,
    max_number_of_workers: int = 10,
    ssl_server_cert: Optional[bytes] = None,
    ssl_server_cert_key: Optional[bytes] = None,
    ssl_ca_cert: Optional[bytes] = None,
    auto_reload: bool = False,
    endpoints: str = DEFAULT_ENDPOINTS_PATH,
) -> grpc.Server:
    """Create a gRPC server to handle incoming action requests.

    Args:
        action_executor: The action executor.
        port: Port to start the server on.
        max_number_of_workers: Maximum number of workers to use.
        ssl_server_cert: File path to the SSL certificate.
        ssl_server_cert_key: File path to the SSL key file.
        ssl_ca_cert: File path to the SSL CA certificate file.
        auto_reload: Enable auto-reloading of modules containing Action subclasses.
        endpoints: Path to the endpoints file.

    Returns:
        The gRPC server.
    """
    server = aio.server(
        futures.ThreadPoolExecutor(max_workers=max_number_of_workers),
        compression=grpc.Compression.Gzip,
    )

    _initialise_health_service(server)
    _initialise_action_service(server, action_executor, auto_reload, endpoints)
    _initialise_port(server, port, ssl_server_cert, ssl_server_cert_key, ssl_ca_cert)

    return server


async def run_grpc(
    action_executor: ActionExecutor,
    port: int = DEFAULT_SERVER_PORT,
    ssl_server_cert_path: Optional[str] = None,
    ssl_server_cert_key_file_path: Optional[str] = None,
    ssl_ca_file_path: Optional[str] = None,
    auto_reload: bool = False,
    endpoints: str = DEFAULT_ENDPOINTS_PATH,
):
    """Start a gRPC server to handle incoming action requests.

    Args:
        action_executor: The action executor.
        port: Port to start the server on.
        ssl_server_cert_path: File path to the client SSL certificate.
        ssl_server_cert_key_file_path: File path to the SSL key for client cert.
        ssl_ca_file_path: File path to the SSL CA certificate file.
        auto_reload: Enable auto-reloading of modules containing Action subclasses.
        endpoints: Path to the endpoints file.
    """
    max_number_of_workers = number_of_sanic_workers()
    ssl_server_cert = (
        file_as_bytes(ssl_server_cert_path) if (ssl_server_cert_path) else None
    )
    ssl_server_cert_key = (
        file_as_bytes(ssl_server_cert_key_file_path)
        if (ssl_server_cert_key_file_path)
        else None
    )
    ssl_ca_cert = file_as_bytes(ssl_ca_file_path) if (ssl_ca_file_path) else None

    server = _initialise_grpc_server(
        action_executor,
        port,
        max_number_of_workers,
        ssl_server_cert,
        ssl_server_cert_key,
        ssl_ca_cert,
        auto_reload,
        endpoints,
    )

    _initialise_interrupts(server)

    await server.start()
    logger.info(f"gRPC Server started on port {port}")
    await server.wait_for_termination()



import argparse
import logging
import os
import warnings
import zlib
import json
from functools import partial
from typing import List, Text, Union, Optional, Any
from ssl import SSLContext

from multidict import MultiDict
from sanic import Sanic, response
from sanic.compat import Header
from sanic.response import HTTPResponse
from sanic.worker.loader import AppLoader

# catching:
# - all `pkg_resources` deprecation warning from multiple dependencies
# - google rcp warnings (`pkg_resources.namespaces`)
# - open telemetry (`pkg_resources`)
# - sanic-cors (`distutils Version classes...`)
with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore", category=DeprecationWarning, message=".*pkg_resources.*"
    )
    warnings.filterwarnings(
        "ignore",
        category=DeprecationWarning,
        message="distutils Version classes are deprecated",
    )
    from sanic_cors import CORS
    from sanic.request import Request
    from rasa_sdk import utils
    from rasa_sdk.cli.arguments import add_endpoint_arguments
    from rasa_sdk.constants import (
        DEFAULT_ENDPOINTS_PATH,
        DEFAULT_KEEP_ALIVE_TIMEOUT,
        DEFAULT_SERVER_PORT,
    )
    from rasa_sdk.executor import ActionExecutor
    from rasa_sdk.interfaces import (
        ActionExecutionRejection,
        ActionNotFoundException,
        ActionMissingDomainException,
    )
    from rasa_sdk.plugin import plugin_manager
    from rasa_sdk.tracing.utils import (
        get_tracer_and_context,
        get_tracer_provider,
        set_span_attributes,
    )

logger = logging.getLogger(__name__)


def configure_cors(
    app: Sanic, cors_origins: Union[Text, List[Text], None] = ""
) -> None:
    """Configure CORS origins for the given app."""
    CORS(
        app, resources={r"/*": {"origins": cors_origins or ""}}, automatic_options=True
    )


def create_ssl_context(
    ssl_certificate: Optional[Text],
    ssl_keyfile: Optional[Text],
    ssl_password: Optional[Text] = None,
) -> Optional[SSLContext]:
    """Create a SSL context if a certificate is passed."""
    if ssl_certificate:
        import ssl

        ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(
            ssl_certificate, keyfile=ssl_keyfile, password=ssl_password
        )
        return ssl_context
    else:
        return None


def create_argument_parser():
    """Parse all the command line arguments for the run script."""
    parser = argparse.ArgumentParser(description="starts the action endpoint")
    add_endpoint_arguments(parser)
    utils.add_logging_level_option_arguments(parser)
    utils.add_logging_file_arguments(parser)
    return parser


async def load_tracer_provider(endpoints: str, app: Sanic):
    """Load the tracer provider into the Sanic app."""
    tracer_provider = get_tracer_provider(endpoints)
    app.ctx.tracer_provider = tracer_provider


def create_app(
    action_executor: ActionExecutor,
    cors_origins: Union[Text, List[Text], None] = "*",
    auto_reload: bool = False,
) -> Sanic:
    """Create a Sanic application and return it.

    Args:
        action_executor: The action executor to use.
        cors_origins: CORS origins to allow.
        auto_reload: When `True`, auto-reloading of actions is enabled.

    Returns:
        A new Sanic application ready to be run.
    """
    app = Sanic("rasa_sdk", configure_logging=False)

    # Reset Sanic warnings filter that allows the triggering of Sanic warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning, module=r"sanic.*")

    configure_cors(app, cors_origins)

    app.ctx.tracer_provider = None

    @app.get("/health")
    async def health(_) -> HTTPResponse:
        """Ping endpoint to check if the server is running and well."""
        body = {"status": "ok"}
        return response.json(body, status=200)

    @app.post("/webhook")
    async def webhook(request: Request) -> HTTPResponse:
        """Webhook to retrieve action calls."""
        span_name = "create_app.webhook"

        def header_to_multi_dict(headers: Header) -> MultiDict:
            return MultiDict(
                [
                    (key, value)
                    for key, value in headers.items()
                    if key.lower() not in ("content-length", "content-encoding")
                ]
            )

        tracer, context = get_tracer_and_context(
            span_name=span_name,
            tracer_provider=request.app.ctx.tracer_provider,
            tracing_carrier=header_to_multi_dict(request.headers),
        )

        with tracer.start_as_current_span(span_name, context=context) as span:
            if request.headers.get("Content-Encoding") == "deflate":
                # Decompress the request data using zlib
                decompressed_data = zlib.decompress(request.body)
                # Load the JSON data from the decompressed request data
                action_call = json.loads(decompressed_data)
            else:
                action_call = request.json
            if action_call is None:
                body = {"error": "Invalid body request"}
                return response.json(body, status=400)

            utils.check_version_compatibility(action_call.get("version"))

            if auto_reload:
                action_executor.reload()
            try:
                result = await action_executor.run(action_call)
            except ActionExecutionRejection as e:
                logger.debug(e)
                body = {"error": e.message, "action_name": e.action_name}
                return response.json(body, status=400)
            except ActionNotFoundException as e:
                logger.error(e)
                body = {"error": e.message, "action_name": e.action_name}
                return response.json(body, status=404)
            except ActionMissingDomainException as e:
                logger.debug(e)
                body = {"error": e.message, "action_name": e.action_name}
                return response.json(body, status=449)

            set_http_span_attributes(
                span,
                action_call,
                http_method="POST",
                route="/webhook",
            )

            return response.json(
                result.model_dump() if result else None,
                status=200,
            )

    @app.get("/actions")
    async def actions(_) -> HTTPResponse:
        """List all registered actions."""
        if auto_reload:
            action_executor.reload()

        body = [
            action_name_item.model_dump()
            for action_name_item in action_executor.list_actions()
        ]
        return response.json(body, status=200)

    @app.exception(Exception)
    async def exception_handler(request, exception: Exception):
        logger.error(
            msg=f"Exception occurred during execution of request {request}",
            exc_info=exception,
        )
        body = {"error": str(exception), "request_body": request.json}
        return response.json(body, status=500)

    return app


def run(
    action_executor: ActionExecutor,
    port: int = DEFAULT_SERVER_PORT,
    cors_origins: Union[Text, List[Text], None] = "*",
    ssl_certificate: Optional[Text] = None,
    ssl_keyfile: Optional[Text] = None,
    ssl_password: Optional[Text] = None,
    auto_reload: bool = False,
    endpoints: str = DEFAULT_ENDPOINTS_PATH,
    keep_alive_timeout: int = DEFAULT_KEEP_ALIVE_TIMEOUT,
) -> None:
    """Starts the action endpoint server with given config values."""
    logger.info("Starting action endpoint server...")
    loader = AppLoader(
        factory=partial(
            create_app,
            action_executor,
            cors_origins=cors_origins,
            auto_reload=auto_reload,
        ),
    )
    app = loader.load()

    app.config.KEEP_ALIVE_TIMEOUT = keep_alive_timeout

    app.register_listener(
        partial(load_tracer_provider, endpoints),
        "before_server_start",
    )

    # Attach additional sanic extensions: listeners, middleware and routing
    logger.info("Starting plugins...")
    plugin_manager().hook.attach_sanic_app_extensions(app=app)

    ssl_context = create_ssl_context(ssl_certificate, ssl_keyfile, ssl_password)
    protocol = "https" if ssl_context else "http"
    host = os.environ.get("SANIC_HOST", "0.0.0.0")

    logger.info(f"Action endpoint is up and running on {protocol}://{host}:{port}")
    app.run(
        host=host,
        port=port,
        ssl=ssl_context,
        workers=utils.number_of_sanic_workers(),
        legacy=True,
    )


def set_http_span_attributes(
    span: Any,
    action_call: dict,
    http_method: str,
    route: str,
) -> None:
    """Sets http span attributes."""
    set_span_attributes(span, action_call)
    if span.is_recording():
        span.set_attribute("http.method", http_method)
        span.set_attribute("http.route", route)


if __name__ == "__main__":
    import rasa_sdk.__main__

    rasa_sdk.__main__.main()



import json

from typing import Any, Dict, Text, List
from rasa_sdk.executor import ActionExecutor, CollectingDispatcher
from rasa_sdk.forms import ValidationAction
from rasa_sdk.types import ActionCall, DomainDict
from rasa_sdk import Tracker


# This file contains all attribute extractors for tracing instrumentation.
# These are functions that are applied to the arguments of the wrapped function to be
# traced to extract the attributes that we want to forward to our tracing backend.
# Note that we always mirror the argument lists of the wrapped functions, as our
# wrapping mechanism always passes in the original arguments unchanged for further
# processing.


def extract_attrs_for_action_executor(
    self: ActionExecutor,
    action_call: ActionCall,
) -> Dict[Text, Any]:
    """Extract the attributes for `ActionExecutor.run`.

    :param self: The `ActionExecutor` on which `run` is called.
    :param action_call: The `ActionCall` argument.
    :return: A dictionary containing the attributes.
    """
    attributes = {"sender_id": action_call.get("sender_id", "None")}
    action_name = action_call.get("next_action")

    if action_name:
        attributes["action_name"] = action_name

    return attributes


def extract_attrs_for_validation_action(
    self: ValidationAction,
    dispatcher: CollectingDispatcher,
    tracker: Tracker,
    domain: DomainDict,
) -> Dict[Text, Any]:
    """Extract the attributes for `ValidationAction.run`.

    :param self: The `ValidationAction` on which `run` is called.
    :param dispatcher: The `CollectingDispatcher` argument.
    :param tracker: The `Tracker` argument.
    :param domain: The `DomainDict` argument.
    :return: A dictionary containing the attributes.
    """
    slots_to_validate = tracker.slots_to_validate().keys()

    return {
        "class_name": self.__class__.__name__,
        "sender_id": tracker.sender_id,
        "slots_to_validate": json.dumps(list(slots_to_validate)),
        "action_name": self.name(),
    }


def extract_attrs_for_action_executor_create_api_response(
    events: List[Dict[Text, Any]],
    messages: List[Dict[Text, Any]],
) -> Dict[Text, Any]:
    """Extract the attributes for `ActionExecutor.run`.

    :param events: A list of events.
    :param messsages: A list of bot responses.
    :return: A dictionary containing the attributes.
    """
    event_names = []
    slot_names = []

    for event in events:
        event_names.append(event.get("event"))
        if event.get("event") == "slot" and event.get("name") != "requested_slot":
            slot_names.append(event.get("name"))
    utters = [
        message.get("response") for message in messages if message.get("response")
    ]

    return {
        "events": json.dumps(list(dict.fromkeys(event_names))),
        "slots": json.dumps(list(dict.fromkeys(slot_names))),
        "utters": json.dumps(utters),
        "message_count": len(messages),
    }



from __future__ import annotations

import abc
import logging
import os
from typing import Any, Dict, Optional, Text

import grpc
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from rasa_sdk.tracing.endpoints import EndpointConfig, read_endpoint_config
from rasa_sdk.tracing.instrumentation import instrumentation
from rasa_sdk.executor import ActionExecutor
from rasa_sdk.forms import ValidationAction, FormValidationAction

TRACING_SERVICE_NAME = os.environ.get("RASA_SDK_TRACING_SERVICE_NAME", "rasa_sdk")

ENDPOINTS_TRACING_KEY = "tracing"
ENDPOINTS_TRACING_SERVICE_NAME_KEY = "service_name"

logger = logging.getLogger(__name__)


def configure_tracing(tracer_provider: Optional[TracerProvider]) -> None:
    """Configure tracing functionality.

    When a tracing backend is defined, this function will
    instrument all methods that shall be traced.
    If no tracing backend is defined, no tracing is configured.

    :param tracer_provider: The `TracingProvider` to be used for tracing
    """
    if tracer_provider is None:
        return None

    instrumentation.instrument(
        tracer_provider=tracer_provider,
        action_executor_class=ActionExecutor,
        validation_action_class=ValidationAction,
        form_validation_action_class=FormValidationAction,
    )


def get_tracer_provider(endpoints_file: Text) -> Optional[TracerProvider]:
    """Configure tracing backend.

    When a known tracing backend is defined in the endpoints file, this
    function will configure the tracing infrastructure. When no or an unknown
    tracing backend is defined, this function does nothing.

    :param endpoints_file: The configuration file containing information about the
        tracing backend.
    :return: The `TracingProvider` to be used for all subsequent tracing.
    """
    cfg = read_endpoint_config(endpoints_file, ENDPOINTS_TRACING_KEY)

    if not cfg:
        logger.info(
            f"No endpoint for tracing type available in {endpoints_file},"
            f"tracing will not be configured."
        )
        return None
    if cfg.type == "jaeger":
        tracer_provider = JaegerTracerConfigurer.configure_from_endpoint_config(cfg)
    elif cfg.type == "otlp":
        tracer_provider = OTLPCollectorConfigurer.configure_from_endpoint_config(cfg)
    else:
        logger.warning(
            f"Unknown tracing type {cfg.type} read from {endpoints_file}, ignoring."
        )
        return None

    return tracer_provider


class TracerConfigurer(abc.ABC):
    """Abstract superclass for tracing configuration.

    `TracerConfigurer` is the abstract superclass from which all configurers
    for different supported backends should inherit.
    """

    @classmethod
    @abc.abstractmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing.

        This abstract method should be implemented by all concrete `TracerConfigurer`s.
        It shall read the configuration from the supplied argument, configure all
        necessary infrastructure for tracing, and return the `TracerProvider` to be
        used for tracing purposes.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """


class JaegerTracerConfigurer(TracerConfigurer):
    """The `TracerConfigurer` for a Jaeger backend.

    This class maintains backward compatibility with the old Jaeger configuration format
    while internally using OTLP to avoid protobuf compatibility issues.
    """

    @classmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing for Jaeger using OTLP under the hood.

        This maintains backward compatibility with the old Jaeger configuration format
        while using OTLP internally to avoid protobuf compatibility issues.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """
        # Extract Jaeger-specific configuration
        jaeger_config = cls._extract_config(cfg)

        # Map Jaeger configuration to OTLP endpoint
        otlp_endpoint = f"http://{jaeger_config['agent_host_name']}:{jaeger_config.get('agent_port', 4317)}"  # noqa: E501

        # Create OTLP exporter with Jaeger-compatible configuration
        otlp_exporter = OTLPSpanExporter(
            endpoint=otlp_endpoint,
            insecure=True,  # Jaeger typically runs without TLS in development
            headers=cls._build_headers(jaeger_config),
        )

        provider = TracerProvider(
            resource=Resource.create(
                {
                    SERVICE_NAME: cfg.kwargs.get(
                        ENDPOINTS_TRACING_SERVICE_NAME_KEY, TRACING_SERVICE_NAME
                    )
                }
            )
        )

        logger.info(
            f"Registered {cfg.type} endpoint for tracing using OTLP. "
            f"Traces will be exported to {otlp_endpoint}."
        )
        provider.add_span_processor(BatchSpanProcessor(otlp_exporter))

        return provider

    @classmethod
    def _extract_config(cls, cfg: EndpointConfig) -> Dict[str, Any]:
        """Extract Jaeger configuration parameters."""
        return {
            "agent_host_name": (cfg.kwargs.get("host", "localhost")),
            "agent_port": (cfg.kwargs.get("port", 6831)),
            "username": cfg.kwargs.get("username"),
            "password": cfg.kwargs.get("password"),
        }

    @classmethod
    def _build_headers(
        cls, jaeger_config: Dict[str, Any]
    ) -> Optional[list[tuple[str, str]]]:
        """Build OTLP headers from Jaeger authentication config."""
        headers: list[tuple[str, str]] = []
        if jaeger_config.get("username") and jaeger_config.get("password"):
            import base64

            credentials = base64.b64encode(
                f"{jaeger_config['username']}:{jaeger_config['password']}".encode()
            ).decode()
            headers.append(("Authorization", f"Basic {credentials}"))
        return headers if headers else None


class OTLPCollectorConfigurer(TracerConfigurer):
    """The `TracerConfigurer` for an OTLP collector backend."""

    @classmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing for Jaeger.

        This will read the OTLP collector-specific configuration from the
        `EndpointConfig` and create a corresponding `TracerProvider` that exports to
        the given OTLP collector.
        Currently, this only supports insecure connections via gRPC.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """
        provider = TracerProvider(
            resource=Resource.create({SERVICE_NAME: TRACING_SERVICE_NAME})
        )

        insecure = cfg.kwargs.get("insecure")

        credentials = cls._get_credentials(cfg, insecure)  # type: ignore

        otlp_exporter = OTLPSpanExporter(
            endpoint=cfg.kwargs["endpoint"],
            insecure=insecure,
            credentials=credentials,
        )
        logger.info(
            f"Registered {cfg.type} endpoint for tracing."
            f"Traces will be exported to {cfg.kwargs['endpoint']}"
        )
        provider.add_span_processor(BatchSpanProcessor(otlp_exporter))

        return provider

    @classmethod
    def _get_credentials(
        cls, cfg: EndpointConfig, insecure: bool
    ) -> Optional[grpc.ChannelCredentials]:
        credentials = None
        if not insecure and "root_certificates" in cfg.kwargs:
            with open(cfg.kwargs.get("root_certificates"), "rb") as f:  # type: ignore
                root_cert = f.read()
            credentials = grpc.ssl_channel_credentials(root_certificates=root_cert)
        return credentials



from typing import Optional
from rasa_sdk.utils import Singleton
from opentelemetry.trace import Tracer


class ActionExecutorTracerRegister(metaclass=Singleton):
    """Represents a provider for ActionExecutor tracer."""

    tracer: Optional[Tracer] = None

    def register_tracer(self, tracer: Tracer) -> None:
        """Register an ActionExecutor tracer.

        Args:
            tracer: The tracer to register.
        """
        self.tracer = tracer

    def get_tracer(self) -> Optional[Tracer]:
        """Get the ActionExecutor tracer.

        Returns:
            The tracer.
        """
        return self.tracer



from typing import Any, Dict, List, Optional, Text

from typing_extensions import TypedDict


class TrackerState(TypedDict):
    """A dictionary representation of the state of a conversation."""

    # id of the source of the messages
    sender_id: Text
    # the currently set values of the slots
    slots: Dict[Text, Any]
    # the most recent message sent by the user
    latest_message: Optional[Dict[Text, Any]]
    # list of previously seen events
    events: List[Dict[Text, Any]]
    # whether the tracker is currently paused
    paused: bool
    # a deterministically scheduled action to be executed next
    followup_action: Optional[Text]
    # the loop that is currently active
    active_loop: Dict[Text, Any]
    # `active_form` is deprecated in favor of `active_loop`
    active_form: Dict[Text, Any]
    # the name of the previously executed action or text of e2e action
    latest_action_name: Optional[Text]
    # the current dialogue stack
    stack: List[Dict[Text, Any]]
    # optional user id
    user_id: Optional[Text]


class DomainDict(TypedDict):
    """A dictionary representation of the domain."""

    intents: List[Dict[Text, Any]]
    entities: List[Text]
    slots: Dict[Any, Dict[Text, Any]]
    responses: Dict[Text, List[Dict[Text, Any]]]
    actions: List[Text]
    forms: Dict[Text, Any]


class ActionCall(TypedDict):
    """A dictionary representation of an action to be executed."""

    # the name of the next action to be executed
    next_action: Optional[Text]
    # id of the source of the messages
    sender_id: Text
    # current dictionary representation of the state of a conversation
    tracker: TrackerState
    # dictionary representation of the domain
    domain: DomainDict
    # rasa version
    version: Text



from __future__ import annotations

from pydantic import BaseModel, Field

from enum import Enum


class ActionExecutionFailed(BaseModel):
    """Error which indicates that an action execution failed.

    Attributes:
        action_name: Name of the action that failed.
        message: Message which describes the error.
    """

    action_name: str = Field(alias="action_name")
    message: str = Field(alias="message")


class ResourceNotFoundType(str, Enum):
    """Type of resource that was not found."""

    ACTION = "ACTION"
    DOMAIN = "DOMAIN"


class ResourceNotFound(BaseModel):
    """Error which indicates that a resource was not found.

    Attributes:
        action_name: Name of the action that was not found.
        message: Message which describes the error.
    """

    action_name: str = Field(alias="action_name")
    message: str = Field(alias="message")
    resource_type: ResourceNotFoundType = Field(alias="resource_type")



import copy
import logging
import typing
import warnings
from typing import Any, Dict, Iterator, List, Optional, Text

from rasa_sdk.events import EventType

if typing.TYPE_CHECKING:  # pragma: no cover
    from rasa_sdk.executor import CollectingDispatcher
    from rasa_sdk.types import DomainDict, TrackerState


logger = logging.getLogger(__name__)

ACTION_LISTEN_NAME = "action_listen"
NLU_FALLBACK_INTENT_NAME = "nlu_fallback"


class Tracker:
    """Maintains the state of a conversation."""

    @classmethod
    def from_dict(cls, state: "TrackerState") -> "Tracker":
        """Create a tracker from dump."""
        return Tracker(
            state["sender_id"],
            state.get("slots", {}),
            state.get("latest_message", {}),
            state.get("events", []),
            state.get("paused", False),
            state.get("followup_action"),
            state.get("active_loop", state.get("active_form", {})),
            state.get("latest_action_name"),
            state.get("stack", []),
            state.get("user_id"),
        )

    def __init__(
        self,
        sender_id: Text,
        slots: Dict[Text, Any],
        latest_message: Optional[Dict[Text, Any]],
        events: List[Dict[Text, Any]],
        paused: bool,
        followup_action: Optional[Text],
        active_loop: Dict[Text, Any],
        latest_action_name: Optional[Text],
        stack: Optional[List[Dict[Text, Any]]] = None,
        user_id: Optional[Text] = None,
    ) -> None:
        """Initialize the tracker."""
        # list of previously seen events
        self.events = events
        # id of the source of the messages
        self.sender_id = sender_id
        # slots that can be filled in this domain
        self.slots = slots

        self.followup_action = followup_action

        self._paused = paused

        # latest_message is `parse_data`,
        # which is a dict: {"intent": UserUttered.intent,
        #                   "entities": UserUttered.entities,
        #                   "text": text}
        self.latest_message = latest_message if latest_message else {}
        self.active_loop = active_loop
        self.latest_action_name = latest_action_name
        self.stack = stack if stack else []
        self.user_id = user_id

    @property
    def active_form(self) -> Dict[Text, Any]:
        """Get the currently active form."""
        warnings.warn(
            "Use of `active_form` is deprecated. Please use `active_loop insteaad.",
            DeprecationWarning,
        )
        return self.active_loop

    def current_state(self) -> Dict[Text, Any]:
        """Return the current tracker state as an object."""
        if len(self.events) > 0:
            latest_event_time = self.events[-1].get("timestamp")
        else:
            latest_event_time = None

        return {
            "sender_id": self.sender_id,
            "slots": self.slots,
            "latest_message": self.latest_message,
            "latest_event_time": latest_event_time,
            "paused": self.is_paused(),
            "events": self.events,
            "latest_input_channel": self.get_latest_input_channel(),
            "active_loop": self.active_loop,
            "latest_action_name": self.latest_action_name,
            "stack": self.stack,
            "user_id": self.user_id,
        }

    def current_slot_values(self) -> Dict[Text, Any]:
        """Return the currently set values of the slots."""
        return self.slots

    def get_slot(self, key) -> Optional[Any]:
        """Retrieves the value of a slot."""
        if key in self.slots:
            return self.slots[key]
        else:
            logger.info(f"Tried to access non existent slot '{key}'.")
            return None

    def get_latest_entity_values(
        self,
        entity_type: Text,
        entity_role: Optional[Text] = None,
        entity_group: Optional[Text] = None,
    ) -> Iterator[Text]:
        """Get entity values found for the passed entity type.

        Optionally role and group of the entities in the last message can be specified.

        If you are only interested in the first entity of a given type use
        `next(tracker.get_latest_entity_values("my_entity_name"), None)`.
        If no entity is found `None` is the default result.

        Args:
            entity_type: the entity type of interest
            entity_role: optional entity role of interest
            entity_group: optional entity group of interest

        Returns:
            List of entity values.
        """
        entities = self.latest_message.get("entities", [])
        return (
            x.get("value")
            for x in entities
            if x.get("entity") == entity_type
            and x.get("group") == entity_group
            and x.get("role") == entity_role
        )

    def get_latest_input_channel(self) -> Optional[Text]:
        """Get the name of the input_channel of the latest UserUttered event."""
        for e in reversed(self.events):
            if e.get("event") == "user":
                return e.get("input_channel")
        return None

    def is_paused(self) -> bool:
        """State whether the tracker is currently paused."""
        return self._paused

    def idx_after_latest_restart(self) -> int:
        """Return the idx of the most recent restart in the list of events.

        If the conversation has not been restarted, `0` is returned.
        """
        idx = 0
        for i, event in enumerate(self.events):
            if event.get("event") == "restart":
                idx = i + 1
        return idx

    def events_after_latest_restart(self) -> List[dict]:
        """Return a list of events after the most recent restart."""
        return list(self.events)[self.idx_after_latest_restart() :]

    @property
    def active_loop_name(self) -> Optional[Text]:
        """Get the name of the currently active loop.

        Returns: `None` if no active loop or the name of the currently active loop.
        """
        if not self.active_loop or self.active_loop.get("name") == "should_not_be_set":
            return None

        return self.active_loop.get("name")

    def __eq__(self, other: Any) -> bool:
        if isinstance(self, type(other)):
            return other.events == self.events and self.sender_id == other.sender_id
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def copy(self) -> "Tracker":
        return Tracker(
            self.sender_id,
            copy.deepcopy(self.slots),
            copy.deepcopy(self.latest_message),
            copy.deepcopy(self.events),
            self._paused,
            self.followup_action,
            self.active_loop,
            self.latest_action_name,
            self.stack,
            self.user_id,
        )

    def last_executed_action_has(self, name: Text, skip: int = 0) -> bool:
        last = self.get_last_event_for(
            "action", exclude=[ACTION_LISTEN_NAME], skip=skip
        )
        return last is not None and last["name"] == name

    def get_last_event_for(
        self, event_type: Text, exclude: List[Text] = [], skip: int = 0
    ) -> Optional[Dict[Text, Any]]:
        def filter_function(e: Dict[Text, Any]) -> bool:
            has_instance = e["event"] == event_type
            excluded = e["event"] == "action" and e["name"] in exclude

            return has_instance and not excluded

        filtered = filter(filter_function, reversed(self.applied_events()))
        for _ in range(skip):
            next(filtered, None)

        return next(filtered, None)

    def applied_events(self) -> List[Dict[Text, Any]]:
        """Returns all actions that should be applied - w/o reverted events."""

        def undo_till_previous(
            event_type: Text, done_events: List[Dict[Text, Any]]
        ) -> None:
            """Removes events from `done_events` until `event_type` is found.

            Removes all events until first occurrence of an `event_type` is found
            including the `event_type`.

            Args:
                event_type: The type of event to remove.
                done_events: The list of events to remove the event from.
            """
            # list gets modified - hence we need to copy events!
            for e in reversed(done_events[:]):
                del done_events[-1]
                if e["event"] == event_type:
                    break

        applied_events: List[Dict[Text, Any]] = []
        for event in self.events:
            event_type = event.get("event")
            if event_type == "restart":
                applied_events = []
            elif event_type == "undo":
                undo_till_previous("action", applied_events)
            elif event_type == "rewind":
                # Seeing a user uttered event automatically implies there was
                # a listen event right before it, so we'll first rewind the
                # user utterance, then get the action right before it (also removes
                # the `action_listen` action right before it).
                undo_till_previous("user", applied_events)
                undo_till_previous("action", applied_events)
            else:
                applied_events.append(event)
        return applied_events

    def slots_to_validate(self) -> Dict[Text, Any]:
        """Get slots which were recently set.

        This can e.g. be used to validate form slots after they were extracted.

        Returns:
            A mapping of extracted slot candidates and their values.
        """
        slots: Dict[Text, Any] = {}
        count: int = 0

        for event in reversed(self.events):
            # The `FormAction` in Rasa Open Source will append all slot candidates
            # at the end of the tracker events.
            if event["event"] == "slot":
                count += 1
            else:
                # Stop as soon as there is another event type as this means that we
                # checked all potential slot candidates.
                break

        for event in self.events[len(self.events) - count :]:
            slots[event["name"]] = event["value"]

        return slots

    def add_slots(self, slots: List[EventType]) -> None:
        """Adds slots to the current tracker.

        Args:
            slots: `SlotSet` events.
        """
        for event in slots:
            if not event.get("event") == "slot":
                continue
            self.slots[event["name"]] = event["value"]
            self.events.append(event)

    def get_intent_of_latest_message(
        self, skip_fallback_intent: bool = True
    ) -> Optional[Text]:
        """Retrieves the intent the last user message.

        Args:
            skip_fallback_intent: Optionally skip the nlu_fallback intent
                and return the next.

        Returns:
            Intent of latest message if available.
        """
        latest_message = self.latest_message
        if not latest_message:
            return None

        intent_ranking = latest_message.get("intent_ranking")
        if not intent_ranking:
            return None

        highest_ranking_intent = intent_ranking[0]
        if (
            highest_ranking_intent["name"] == NLU_FALLBACK_INTENT_NAME
            and skip_fallback_intent
        ):
            if len(intent_ranking) >= 2:
                return intent_ranking[1]["name"]
            else:
                return None
        else:
            return highest_ranking_intent["name"]


class Action:
    """Next action to be taken in response to a dialogue state."""

    def name(self) -> Text:
        """Unique identifier of this simple action."""
        raise NotImplementedError("An action must implement a name")

    async def run(
        self,
        dispatcher: "CollectingDispatcher",
        tracker: Tracker,
        domain: "DomainDict",
    ) -> List[Dict[Text, Any]]:
        """Execute the side effects of this action.

        Args:
            dispatcher: the dispatcher which is used to
                send messages back to the user. Use
                `dispatcher.utter_message()` for sending messages.
            tracker: the state tracker for the current
                user. You can access slot values using
                `tracker.get_slot(slot_name)`, the most recent user message
                is `tracker.latest_message.text` and any other
                `rasa_sdk.Tracker` property.
            domain: the bot's domain
        Returns:
            A dictionary of `rasa_sdk.events.Event` instances that is
                returned through the endpoint
        """
        raise NotImplementedError("An action must implement its run method")

    def __str__(self) -> Text:
        return f"Action('{self.name()}')"


class ActionExecutionRejection(Exception):
    """Raising this exception will allow other policies to predict another action."""

    def __init__(self, action_name: Text, message: Optional[Text] = None) -> None:
        """Create a rejection exception.

        Args:
            action_name: Name of the action that should be rejected.
            message: Optional message to provide more information
        """
        self.action_name = action_name
        self.message = message or f"Custom action '{action_name}' rejected execution."

    def __str__(self) -> Text:
        """Return the string representation of the exception."""
        return self.message


class ActionNotFoundException(Exception):
    def __init__(self, action_name: Text, message: Optional[Text] = None) -> None:
        """Create an exception for when an action is not found."""
        self.action_name = action_name
        self.message = (
            message or f"No registered action found for name '{action_name}'."
        )

    def __str__(self) -> Text:
        """Return the string representation of the exception."""
        return self.message


class ActionMissingDomainException(Exception):
    """Raising this exception when the domain is missing."""

    def __init__(self, action_name: Text, message: Optional[Text] = None) -> None:
        """Create an exception for when the domain is missing."""
        self.action_name = action_name
        self.message = (
            message
            or "Missing domain context, assistant will retry the request and include "
            "the domain in the request payload. For more information please see "
            "https://rasa.com/docs/action-server/"
        )

    def __str__(self) -> Text:
        """Return the string representation of the exception."""
        return self.message



from __future__ import annotations
import importlib
import inspect
import logging
import pkgutil
import warnings
from typing import Text, List, Dict, Any, Type, Union, Callable, Optional, Set, cast
from collections import namedtuple
import types
import sys
import os

from pydantic import BaseModel, Field

from rasa_sdk.interfaces import (
    Tracker,
    ActionNotFoundException,
    Action,
    ActionMissingDomainException,
)

from rasa_sdk import utils

logger = logging.getLogger(__name__)


class CollectingDispatcher:
    """Send messages back to user."""

    def __init__(self) -> None:
        """Create a `CollectingDispatcher` object."""
        self.messages: List[Dict[Text, Any]] = []

    def utter_message(
        self,
        text: Optional[Text] = None,
        image: Optional[Text] = None,
        json_message: Optional[Dict[Text, Any]] = None,
        template: Optional[Text] = None,
        response: Optional[Text] = None,
        attachment: Optional[Text] = None,
        buttons: Optional[List[Dict[Text, Any]]] = None,
        elements: Optional[List[Dict[Text, Any]]] = None,
        **kwargs: Any,
    ) -> None:
        """Send a text to the output channel."""
        if template and not response:
            response = template
            warnings.warn(
                "Please pass the parameter `response` instead of `template` "
                "to `utter_message`. `template` will be deprecated in Rasa 3.0.0. ",
                FutureWarning,
            )
        message = {
            "text": text,
            "buttons": buttons or [],
            "elements": elements or [],
            "custom": json_message or {},
            "template": response,
            "response": response,
            "image": image,
            "attachment": attachment,
        }
        message.update(kwargs)

        self.messages.append(message)

    # deprecated
    def utter_custom_message(self, *elements: Dict[Text, Any], **kwargs: Any) -> None:
        """Sends a message with custom elements to the output channel.

        Deprecated:
            Use `utter_message(elements=<list of elements>)` instead.

        Args:
            elements: List of elements to be sent to the output channel.
            kwargs: Additional parameters to be sent to the output channel.
        """
        warnings.warn(
            "Use of `utter_custom_message` is deprecated. "
            "Use `utter_message(elements=<list of elements>)` instead.",
            FutureWarning,
        )
        self.utter_message(elements=list(elements), **kwargs)

    def utter_elements(self, *elements: Dict[Text, Any], **kwargs: Any) -> None:
        """Sends a message with custom elements to the output channel."""
        warnings.warn(
            "Use of `utter_elements` is deprecated. "
            "Use `utter_message(elements=<list of elements>)` instead.",
            FutureWarning,
        )
        self.utter_message(elements=list(elements), **kwargs)

    def utter_button_message(
        self, text: Text, buttons: List[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        """Sends a message with buttons to the output channel."""
        warnings.warn(
            "Use of `utter_button_message` is deprecated. "
            "Use `utter_message(text=<text> , buttons=<list of buttons>)` instead.",
            FutureWarning,
        )

        self.utter_message(text=text, buttons=buttons, **kwargs)

    def utter_attachment(self, attachment: Text, **kwargs: Any) -> None:
        """Send a message to the client with attachments."""
        warnings.warn(
            "Use of `utter_attachment` is deprecated. "
            "Use `utter_message(attachment=<attachment>)` instead.",
            FutureWarning,
        )

        self.utter_message(attachment=attachment, **kwargs)

    # noinspection PyUnusedLocal
    def utter_button_template(
        self,
        template: Text,
        buttons: List[Dict[Text, Any]],
        tracker: Tracker,
        silent_fail: bool = False,
        **kwargs: Any,
    ) -> None:
        """Sends a message template with buttons to the output channel."""
        warnings.warn(
            "Use of `utter_button_template` is deprecated. "
            "Use `utter_message(template=<template name>, buttons=<list of buttons>)`"
            " instead.",
            FutureWarning,
        )

        self.utter_message(template=template, buttons=buttons, **kwargs)

    # noinspection PyUnusedLocal
    def utter_template(
        self, template: Text, tracker: Tracker, silent_fail: bool = False, **kwargs: Any
    ) -> None:
        """Send a message to the client based on a template."""
        warnings.warn(
            "Use of `utter_template` is deprecated. "
            "Use `utter_message(response=<template_name>)` instead.",
            FutureWarning,
        )

        self.utter_message(response=template, **kwargs)

    def utter_custom_json(self, json_message: Dict[Text, Any], **kwargs: Any) -> None:
        """Sends custom json to the output channel."""
        warnings.warn(
            "Use of `utter_custom_json` is deprecated. "
            "Use `utter_message(json_message=<message dict>)` instead.",
            FutureWarning,
        )

        self.utter_message(json_message=json_message, **kwargs)

    def utter_image_url(self, image: Text, **kwargs: Any) -> None:
        """Sends url of image attachment to the output channel."""
        warnings.warn(
            "Use of `utter_image_url` is deprecated. "
            "Use `utter_message(image=<image url>)` instead.",
            FutureWarning,
        )

        self.utter_message(image=image, **kwargs)


TimestampModule = namedtuple("TimestampModule", ["timestamp", "module"])


class ActionExecutorRunResult(BaseModel):
    """Model for action executor run result."""

    events: List[Dict[Text, Any]] = Field(alias="events")
    responses: List[Dict[Text, Any]] = Field(alias="responses")


class ActionExecutor:
    """Executes actions."""

    def __init__(self) -> None:
        """Initializes the `ActionExecutor`."""
        self.actions: Dict[Text, Callable] = {}
        self._modules: Dict[Text, TimestampModule] = {}
        self._registered_packages: Set[Text] = set()
        self.domain: Optional[Dict[Text, Any]] = None
        self.domain_digest: Optional[Text] = None

    def register_action(self, action: Union[Type[Action], Action]) -> None:
        """Register an action with the executor.

        Args:
            action: Action to be registered. It can either be an instance of
            `Action` subclass class or an actual `Action` subclass.
        """
        if inspect.isclass(action):
            action = cast(Type[Action], action)
            if action.__module__.startswith("rasa."):
                logger.warning(f"Skipping built in Action {action}.")
                return
            else:
                action = action()

        if isinstance(action, Action):
            self.register_function(action.name(), action.run)
        else:
            raise Exception(
                "You can only register instances or subclasses of "
                "type Action. If you want to directly register "
                "a function, use `register_function` instead."
            )

    def register_function(self, action_name: Text, f: Callable) -> None:
        """Register an executor function for an action.

        Args:
            action_name: Name of the action.
            f: Function to be registered.
        """
        valid_keys = utils.arguments_of(f)
        if len(valid_keys) < 3:
            raise Exception(
                "You can only register functions that take "
                "3 parameters as arguments. The three parameters "
                "your function will receive are: dispatcher, "
                f"tracker, domain. Your function accepts only {len(valid_keys)} "
                "parameters."
            )

        if action_name in self.actions:
            logger.info(f"Re-registered function for '{action_name}'.")
        else:
            logger.info(f"Registered function for '{action_name}'.")

        self.actions[action_name] = f

    def _import_submodules(
        self, package: Union[Text, types.ModuleType], recursive: bool = True
    ) -> None:
        """Import a module, or a package and its submodules recursively.

        Args:
            package: Package or module name, or an already loaded Python module.
            recursive: If `True`, and `package` is a package, import all of its
                sub-packages as well.
        """
        if isinstance(package, str):
            package = self._import_module(package)

        if not getattr(package, "__path__", None):
            return

        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
            full_name = package.__name__ + "." + name
            self._import_module(full_name)

            if recursive and is_pkg:
                self._import_submodules(full_name)

    def _import_module(self, name: Text) -> types.ModuleType:
        """Import a Python module. If possible, register the file where it came from.

        Args:
            name: The module's name using Python's module path syntax.

        Returns:
            The loaded module.
        """
        module = importlib.import_module(name)

        module_file = getattr(module, "__file__", None)
        if module_file:
            # If the module we're importing is a namespace package (a package
            # without __init__.py), then there's nothing to watch for the
            # package itself.
            timestamp = os.path.getmtime(module_file)
            self._modules[module_file] = TimestampModule(timestamp, module)

        return module

    def register_package(self, package: Union[Text, types.ModuleType]) -> None:
        """Register all the `Action` subclasses contained in a Python module or package.

        If an `ImportError` is raised when loading the module or package, the
        action server is stopped with exit code 1.

        Args:
            package: Module or package name, or an already loaded Python module.
        """
        try:
            self._import_submodules(package)
            # Track the package name for reloading
            if package is not None:
                package_name = package if isinstance(package, str) else package.__name__
                self._registered_packages.add(package_name)
        except ImportError:
            logger.exception(f"Failed to register package '{package}'.")
            sys.exit(1)

        self._register_all_actions()

    def _register_all_actions(self) -> None:
        """Scan for all user subclasses of `Action`, and register them."""
        import inspect

        actions = utils.all_subclasses(Action)

        for action in actions:
            if (
                not action.__module__.startswith("rasa_core.")
                and not action.__module__.startswith("rasa.")
                and not action.__module__.startswith("rasa_sdk.")
                and not action.__module__.startswith("rasa_core_sdk.")
                and not inspect.isabstract(action)
            ):
                self.register_action(action)

    def _find_modules_to_reload(self) -> Dict[Text, TimestampModule]:
        """Finds all Python modules that should be reloaded.

         Reloads modules by checking their files' timestamps.

        Returns:
            Dictionary containing file paths, new timestamps and Python modules
            that should be reloaded.
        """
        to_reload = {}

        for path, (timestamp, module) in self._modules.items():
            try:
                new_timestamp = os.path.getmtime(path)
            except OSError:
                # Ignore missing files
                continue

            if new_timestamp > timestamp:
                to_reload[path] = TimestampModule(new_timestamp, module)

        return to_reload

    def reload(self) -> None:
        """Reload all Python modules that have been loaded in the past.

        To check if a module should be reloaded, the file's last timestamp is
        checked against the current one. If the current last-modified timestamp
        is larger than the previous one, the module is re-imported using
        `importlib.reload`.

        Additionally, re-scans registered packages to discover and import any
        newly created modules.

        If one or more modules are reloaded during this process, the entire
        `Action` class hierarchy is scanned again to see what new classes can
        be registered.
        """
        to_reload = self._find_modules_to_reload()
        any_module_reloaded = False

        # Reload modified modules
        for path, (timestamp, module) in to_reload.items():
            try:
                new_module = importlib.reload(module)
                self._modules[path] = TimestampModule(timestamp, new_module)
                logger.info(
                    f"Reloaded module/package: '{module.__name__}' (file: '{path}')"
                )
                any_module_reloaded = True
            except (SyntaxError, ImportError):
                logger.exception(
                    f"Error while reloading module/package: '{module.__name__}' "
                    f"(file: '{path}'):"
                )
                logger.info("Please fix the error(s) in the Python file and try again.")

        # Re-scan registered packages to discover new modules
        for package_name in self._registered_packages:
            try:
                # Get current module files before re-import
                old_module_files = set(self._modules.keys())
                self._import_submodules(package_name)
                # Check if any new modules were imported
                new_module_files = set(self._modules.keys())
                if new_module_files != old_module_files:
                    newly_imported = new_module_files - old_module_files
                    for module_file in newly_imported:
                        logger.info(f"Discovered new module: '{module_file}'")
                    any_module_reloaded = True
            except ImportError:
                logger.exception(
                    f"Error while scanning package "
                    f"'{package_name}' for new modules:"
                )

        if any_module_reloaded:
            self._register_all_actions()

    @staticmethod
    def _create_api_response(
        events: List[Dict[Text, Any]], messages: List[Dict[Text, Any]]
    ) -> ActionExecutorRunResult:
        return ActionExecutorRunResult(events=events, responses=messages)

    @staticmethod
    def validate_events(
        events: List[Dict[Text, Any]],
        action_name: Text,
    ) -> List[Dict[Text, Any]]:
        """Validate the events returned by the action.

        Args:
            events: List of events returned by the action.

            action_name: Name of the action that should be executed.

        Returns:
            List of validated events.
        """
        validated = []
        for event in events:
            if isinstance(event, dict):
                if not event.get("event"):
                    logger.error(
                        f"Your action '{action_name}' returned an action dict "
                        "without the `event` property. Please use "
                        "the helpers in `rasa_sdk.events`! Event will"
                        f"be ignored!"
                    )
                    logger.debug(f"Contents of ignored event: '{event}'")
                else:
                    validated.append(event)
            elif type(event).__module__ == "rasa.core.events":
                warnings.warn(
                    "Your action should not return Rasa actions within the "
                    "SDK. Instead of using events from "
                    "`rasa.core.events`, you should use the ones "
                    "provided in `rasa_sdk.events`! "
                    "We will try to make this work, but this "
                    "might go wrong!"
                )
                validated.append(event.as_dict())
            else:
                logger.error(
                    f"Your action's '{action_name}' run method returned an invalid "
                    f"event. Event will be ignored."
                )
                logger.debug(f"Contents of ignored event: '{event}'")
                # we won't append this to validated events -> will be ignored
        return validated

    def is_domain_digest_valid(self, domain_digest: Optional[Text]) -> bool:
        """Check if the domain_digest is valid.

        If the domain_digest is empty or different from the one provided, it is invalid.

        Args:
            domain_digest: latest value provided to compare the current value with.

        Returns:
            True if the domain_digest is valid, False otherwise.
        """
        return bool(self.domain_digest) and self.domain_digest == domain_digest

    def update_and_return_domain(
        self, payload: Dict[Text, Any], action_name: Text
    ) -> Optional[Dict[Text, Any]]:
        """Validate the digest, store the domain if available, and return the domain.

        This method validates the domain digest from the payload.
        If the digest is invalid and no domain is provided, an exception is raised.
        If domain data is available, it stores the domain and digest.
        Finally, it returns the domain.

        Args:
            payload: Request payload containing the domain data.
            action_name: Name of the action that should be executed.

        Returns:
            The domain dictionary.

        Raises:
            ActionMissingDomainException: Invalid digest and no domain data available.
        """
        payload_domain = payload.get("domain")
        payload_domain_digest = payload.get("domain_digest")

        # If digest is invalid and no domain is available - raise the error
        if (
            not self.is_domain_digest_valid(payload_domain_digest)
            and payload_domain is None
        ):
            raise ActionMissingDomainException(action_name)

        if payload_domain:
            self.domain = payload_domain
            self.domain_digest = payload_domain_digest

        return self.domain

    async def run(
        self,
        action_call: Dict[Text, Any],
    ) -> Optional[ActionExecutorRunResult]:
        """Run the action and return the response.

        Args:
            action_call: Request payload containing the action data.

        Returns:
            Response containing the events and messages or None if
            the action does not exist.
        """
        from rasa_sdk.interfaces import Tracker

        action_name = action_call.get("next_action")
        if action_name:
            logger.debug(f"Received request to run '{action_name}'")
            action = self.actions.get(action_name)
            if not action:
                raise ActionNotFoundException(action_name)

            tracker_json = action_call["tracker"]
            domain = self.update_and_return_domain(action_call, action_name)
            tracker = Tracker.from_dict(tracker_json)
            dispatcher = CollectingDispatcher()

            events = await utils.call_potential_coroutine(
                action(dispatcher, tracker, domain)
            )

            if not events:
                # make sure the action did not just return `None`...
                events = []

            validated_events = self.validate_events(events, action_name)
            logger.debug(f"Finished running '{action_name}'")
            return self._create_api_response(validated_events, dispatcher.messages)

        logger.warning("Received an action call without an action.")
        return None

    def list_actions(self) -> List[ActionName]:
        """List all registered action names."""
        return [ActionName(name=action_name) for action_name in self.actions.keys()]


class ActionName(BaseModel):
    """Model for action name."""

    name: str = Field(alias="name")



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ActionsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ActionsResponse(_message.Message):
    __slots__ = ("actions",)
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    actions: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    def __init__(self, actions: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ...) -> None: ...

class Tracker(_message.Message):
    __slots__ = ("sender_id", "slots", "latest_message", "events", "paused", "followup_action", "active_loop", "latest_action_name", "stack", "user_id")
    class ActiveLoopEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    SENDER_ID_FIELD_NUMBER: _ClassVar[int]
    SLOTS_FIELD_NUMBER: _ClassVar[int]
    LATEST_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    PAUSED_FIELD_NUMBER: _ClassVar[int]
    FOLLOWUP_ACTION_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_LOOP_FIELD_NUMBER: _ClassVar[int]
    LATEST_ACTION_NAME_FIELD_NUMBER: _ClassVar[int]
    STACK_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    sender_id: str
    slots: _struct_pb2.Struct
    latest_message: _struct_pb2.Struct
    events: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    paused: bool
    followup_action: str
    active_loop: _containers.ScalarMap[str, str]
    latest_action_name: str
    stack: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    user_id: str
    def __init__(self, sender_id: _Optional[str] = ..., slots: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., latest_message: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., events: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ..., paused: bool = ..., followup_action: _Optional[str] = ..., active_loop: _Optional[_Mapping[str, str]] = ..., latest_action_name: _Optional[str] = ..., stack: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ..., user_id: _Optional[str] = ...) -> None: ...

class Intent(_message.Message):
    __slots__ = ("string_value", "dict_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    DICT_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    dict_value: _struct_pb2.Struct
    def __init__(self, string_value: _Optional[str] = ..., dict_value: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class Entity(_message.Message):
    __slots__ = ("string_value", "dict_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    DICT_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    dict_value: _struct_pb2.Struct
    def __init__(self, string_value: _Optional[str] = ..., dict_value: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class Action(_message.Message):
    __slots__ = ("string_value", "dict_value")
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    DICT_VALUE_FIELD_NUMBER: _ClassVar[int]
    string_value: str
    dict_value: _struct_pb2.Struct
    def __init__(self, string_value: _Optional[str] = ..., dict_value: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class Domain(_message.Message):
    __slots__ = ("config", "session_config", "intents", "entities", "slots", "responses", "actions", "forms", "e2e_actions")
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    SESSION_CONFIG_FIELD_NUMBER: _ClassVar[int]
    INTENTS_FIELD_NUMBER: _ClassVar[int]
    ENTITIES_FIELD_NUMBER: _ClassVar[int]
    SLOTS_FIELD_NUMBER: _ClassVar[int]
    RESPONSES_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    FORMS_FIELD_NUMBER: _ClassVar[int]
    E2E_ACTIONS_FIELD_NUMBER: _ClassVar[int]
    config: _struct_pb2.Struct
    session_config: _struct_pb2.Struct
    intents: _containers.RepeatedCompositeFieldContainer[Intent]
    entities: _containers.RepeatedCompositeFieldContainer[Entity]
    slots: _struct_pb2.Struct
    responses: _struct_pb2.Struct
    actions: _containers.RepeatedCompositeFieldContainer[Action]
    forms: _struct_pb2.Struct
    e2e_actions: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    def __init__(self, config: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., session_config: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., intents: _Optional[_Iterable[_Union[Intent, _Mapping]]] = ..., entities: _Optional[_Iterable[_Union[Entity, _Mapping]]] = ..., slots: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., responses: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., actions: _Optional[_Iterable[_Union[Action, _Mapping]]] = ..., forms: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ..., e2e_actions: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ...) -> None: ...

class WebhookRequest(_message.Message):
    __slots__ = ("next_action", "sender_id", "tracker", "domain", "version", "domain_digest")
    NEXT_ACTION_FIELD_NUMBER: _ClassVar[int]
    SENDER_ID_FIELD_NUMBER: _ClassVar[int]
    TRACKER_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_DIGEST_FIELD_NUMBER: _ClassVar[int]
    next_action: str
    sender_id: str
    tracker: Tracker
    domain: Domain
    version: str
    domain_digest: str
    def __init__(self, next_action: _Optional[str] = ..., sender_id: _Optional[str] = ..., tracker: _Optional[_Union[Tracker, _Mapping]] = ..., domain: _Optional[_Union[Domain, _Mapping]] = ..., version: _Optional[str] = ..., domain_digest: _Optional[str] = ...) -> None: ...

class WebhookResponse(_message.Message):
    __slots__ = ("events", "responses")
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    RESPONSES_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    responses: _containers.RepeatedCompositeFieldContainer[_struct_pb2.Struct]
    def __init__(self, events: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ..., responses: _Optional[_Iterable[_Union[_struct_pb2.Struct, _Mapping]]] = ...) -> None: ...



import asyncio
import logging

from rasa_sdk import utils
from rasa_sdk.constants import APPLICATION_ROOT_LOGGER_NAME
from rasa_sdk.endpoint import create_argument_parser, run
from rasa_sdk.executor import ActionExecutor
from rasa_sdk.grpc_server import run_grpc

logger = logging.getLogger(__name__)


def main_from_args(args):
    """Run with arguments."""
    logging.getLogger("matplotlib").setLevel(logging.WARN)

    utils.configure_colored_logging(args.loglevel)
    utils.configure_file_logging(
        logging.getLogger(APPLICATION_ROOT_LOGGER_NAME),
        args.log_file,
        args.loglevel,
        args.logging_config_file,
    )
    utils.update_sanic_log_level()

    action_executor = ActionExecutor()
    action_executor.register_package(
        args.actions_module or args.actions,
    )

    if args.grpc:
        asyncio.run(
            run_grpc(
                action_executor,
                args.port,
                args.ssl_certificate,
                args.ssl_keyfile,
                args.ssl_ca_file,
                args.auto_reload,
                args.endpoints,
            )
        )
    else:
        run(
            action_executor,
            args.port,
            args.cors,
            args.ssl_certificate,
            args.ssl_keyfile,
            args.ssl_password,
            args.auto_reload,
            args.endpoints,
        )


def main():
    """Runs the action server as standalone application."""
    arg_parser = create_argument_parser()
    cmdline_args = arg_parser.parse_args()

    main_from_args(cmdline_args)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: rasa_sdk/grpc_py/action_webhook.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'rasa_sdk/grpc_py/action_webhook.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%rasa_sdk/grpc_py/action_webhook.proto\x12\x15\x61\x63tion_server_webhook\x1a\x1cgoogle/protobuf/struct.proto\"\x10\n\x0e\x41\x63tionsRequest\";\n\x0f\x41\x63tionsResponse\x12(\n\x07\x61\x63tions\x18\x01 \x03(\x0b\x32\x17.google.protobuf.Struct\"\xda\x03\n\x07Tracker\x12\x11\n\tsender_id\x18\x01 \x01(\t\x12&\n\x05slots\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12/\n\x0elatest_message\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\'\n\x06\x65vents\x18\x04 \x03(\x0b\x32\x17.google.protobuf.Struct\x12\x0e\n\x06paused\x18\x05 \x01(\x08\x12\x1c\n\x0f\x66ollowup_action\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x43\n\x0b\x61\x63tive_loop\x18\x07 \x03(\x0b\x32..action_server_webhook.Tracker.ActiveLoopEntry\x12\x1f\n\x12latest_action_name\x18\x08 \x01(\tH\x01\x88\x01\x01\x12&\n\x05stack\x18\t \x03(\x0b\x32\x17.google.protobuf.Struct\x12\x14\n\x07user_id\x18\n \x01(\tH\x02\x88\x01\x01\x1a\x31\n\x0f\x41\x63tiveLoopEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x12\n\x10_followup_actionB\x15\n\x13_latest_action_nameB\n\n\x08_user_id\"K\n\x06Intent\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12+\n\ndict_value\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"K\n\x06\x45ntity\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12+\n\ndict_value\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"K\n\x06\x41\x63tion\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12+\n\ndict_value\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x9d\x03\n\x06\x44omain\x12\'\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12/\n\x0esession_config\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12.\n\x07intents\x18\x03 \x03(\x0b\x32\x1d.action_server_webhook.Intent\x12/\n\x08\x65ntities\x18\x04 \x03(\x0b\x32\x1d.action_server_webhook.Entity\x12&\n\x05slots\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12*\n\tresponses\x18\x06 \x01(\x0b\x32\x17.google.protobuf.Struct\x12.\n\x07\x61\x63tions\x18\x07 \x03(\x0b\x32\x1d.action_server_webhook.Action\x12&\n\x05\x66orms\x18\x08 \x01(\x0b\x32\x17.google.protobuf.Struct\x12,\n\x0b\x65\x32\x65_actions\x18\t \x03(\x0b\x32\x17.google.protobuf.Struct\"\xd7\x01\n\x0eWebhookRequest\x12\x13\n\x0bnext_action\x18\x01 \x01(\t\x12\x11\n\tsender_id\x18\x02 \x01(\t\x12/\n\x07tracker\x18\x03 \x01(\x0b\x32\x1e.action_server_webhook.Tracker\x12-\n\x06\x64omain\x18\x04 \x01(\x0b\x32\x1d.action_server_webhook.Domain\x12\x0f\n\x07version\x18\x05 \x01(\t\x12\x1a\n\rdomain_digest\x18\x06 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_domain_digest\"f\n\x0fWebhookResponse\x12\'\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x17.google.protobuf.Struct\x12*\n\tresponses\x18\x02 \x03(\x0b\x32\x17.google.protobuf.Struct2\xc3\x01\n\rActionService\x12X\n\x07Webhook\x12%.action_server_webhook.WebhookRequest\x1a&.action_server_webhook.WebhookResponse\x12X\n\x07\x41\x63tions\x12%.action_server_webhook.ActionsRequest\x1a&.action_server_webhook.ActionsResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rasa_sdk.grpc_py.action_webhook_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TRACKER_ACTIVELOOPENTRY']._loaded_options = None
  _globals['_TRACKER_ACTIVELOOPENTRY']._serialized_options = b'8\001'
  _globals['_ACTIONSREQUEST']._serialized_start=94
  _globals['_ACTIONSREQUEST']._serialized_end=110
  _globals['_ACTIONSRESPONSE']._serialized_start=112
  _globals['_ACTIONSRESPONSE']._serialized_end=171
  _globals['_TRACKER']._serialized_start=174
  _globals['_TRACKER']._serialized_end=648
  _globals['_TRACKER_ACTIVELOOPENTRY']._serialized_start=544
  _globals['_TRACKER_ACTIVELOOPENTRY']._serialized_end=593
  _globals['_INTENT']._serialized_start=650
  _globals['_INTENT']._serialized_end=725
  _globals['_ENTITY']._serialized_start=727
  _globals['_ENTITY']._serialized_end=802
  _globals['_ACTION']._serialized_start=804
  _globals['_ACTION']._serialized_end=879
  _globals['_DOMAIN']._serialized_start=882
  _globals['_DOMAIN']._serialized_end=1295
  _globals['_WEBHOOKREQUEST']._serialized_start=1298
  _globals['_WEBHOOKREQUEST']._serialized_end=1513
  _globals['_WEBHOOKRESPONSE']._serialized_start=1515
  _globals['_WEBHOOKRESPONSE']._serialized_end=1617
  _globals['_ACTIONSERVICE']._serialized_start=1620
  _globals['_ACTIONSERVICE']._serialized_end=1815
# @@protoc_insertion_point(module_scope)
